import logging
import os
import shutil
import sys
import warnings
import zipfile
from pathlib import Path
from typing import Callable, Dict, List, Union

from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from PIL import Image
from requests import Response
from requests.exceptions import InvalidJSONError

from picsellia import exceptions as exceptions
from picsellia.decorators import exception_handler

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)

try:
    download_bar_mode = os.environ["PICSELLIA_SDK_DOWNLOAD_BAR_MODE"]
except KeyError:
    download_bar_mode = "1"

try:
    print_section_handler = os.environ["PICSELLIA_SDK_SECTION_HANDLER"]
except KeyError:
    print_section_handler = "0"


@exception_handler
@beartype
def zipdir(path):
    zipf = zipfile.ZipFile(path.split(".")[0] + ".zip", "w", zipfile.ZIP_DEFLATED)
    for filepath in os.listdir(path):
        zipf.write(os.path.join(path, filepath), filepath)

        if os.path.isdir(os.path.join(path, filepath)):
            for inside_file in os.listdir(os.path.join(path, filepath)):
                zipf.write(
                    os.path.join(path, filepath, inside_file),
                    os.path.join(filepath, inside_file),
                )

    zipf.close()
    return path.split(".")[0] + ".zip"


@exception_handler
@beartype
def unzip(filename: str, target_dir: Union[str, Path] = "./"):
    shutil.unpack_archive(filename, target_dir)


def handle_response(f):
    def decorated(*args, **kwargs):
        response = f(*args, **kwargs)
        check_status_code(response)
        return response

    return decorated


# No exception handling: exception needs to be raised
def check_status_code(response: Response):
    status = int(response.status_code)
    if status == 200:
        logger.debug("OK.")
    elif status == 201:
        logger.debug("Resource created.")
    elif status == 202:
        logger.debug("Accepted.")
    elif status == 203:
        logger.debug("OK.")
    elif status == 204:
        logger.debug("No content.")
    elif status != 208 and status <= 299:
        logger.debug(f"Request done : {status}")
    else:
        try:
            data = response.json()
            if "message" not in data:
                data["message"] = f"No message ({status})"

            message = data["message"]
            if "detail" in data and data["detail"] is not None and data["detail"] != []:
                if isinstance(data["detail"], list):
                    message += ". Detail: \n"
                    for item in data["detail"]:
                        message += " > " + str(item) + "\n"
                else:
                    message = message + ". Detail: " + str(data["detail"])
        except (KeyError, InvalidJSONError):
            message = response.text

        logger.debug(
            f"Platform returned an error (status code: {status}). Message : {message}"
        )

        if status == 208:
            raise exceptions.DistantStorageError(
                "An object has already this name in S3."
            )
        if status == 400:
            raise exceptions.BadRequestError(message)
        if status == 401:
            raise exceptions.UnauthorizedError(message)
        if status == 402:
            raise exceptions.InsufficientResourcesError(message)
        if status == 403:
            raise exceptions.ForbiddenError(message)
        if status == 404:
            raise exceptions.ResourceNotFoundError(message)
        if status == 405:
            raise exceptions.PicselliaError(f"Method not allowed: {message}")
        if status == 409:
            raise exceptions.ResourceConflictError(message)
        if status == 422:
            raise exceptions.BadRequestError(message)
        if status == 429:
            raise exceptions.TooManyRequestError(message)
        if status == 500:
            raise exceptions.InternalServerError(message)
        if status == 502:
            raise exceptions.BadGatewayError("Picsellia is unavailable at the moment.")
        raise exceptions.PicselliaError(f"[{status}] Something went wrong, {message}")


@exception_handler
@beartype
def generate_requirements_json(requirements_path: Union[str, Path]) -> List[dict]:
    """Generate a json file with the requirements from the requirements.txt file

    Arguments:
        requirements_path ([str]): [absolute path to requirements.txt file]

    Raises:
        exceptions.ResourceNotFoundError: [Filepath does match]
        Exception: [Wrong requirements file]

    Returns:
        [dict]: {
            'requirements': []{
                'package': (str) package name,
                'version': (str) package version
            }
        }
    """
    requirements = []
    try:
        with open(requirements_path, "r") as f:
            lines = f.readlines()
    except Exception:
        raise exceptions.ResourceNotFoundError(f"{requirements_path} does not exists")
    try:
        for line in lines:
            if line[0] != "#":
                try:
                    package, version = line.split("==")
                except Exception:
                    package, version = line, ""
                tmp = {"package": package.rstrip(), "version": version.rstrip()}
                requirements.append(tmp)
    except Exception:  # pragma: no cover
        raise Exception("Malformed requirements file")
    return requirements


def print_line_return():
    if print_section_handler == "1":
        sys.stdout.write("\n")


def print_start_section():
    if print_section_handler == "1":
        sys.stdout.write("-----")


def print_stop_section():
    if print_section_handler == "1":
        sys.stdout.write("--*--")


def print_start_chapter_name(name: str):
    if print_section_handler == "1":
        sys.stdout.write(f"--#--{name}")


def print_logging_buffer(length: int):
    if print_section_handler == "1":
        sys.stdout.write(f"--{length}--")


def filter_payload(payload: dict) -> dict:
    return {k: v for k, v in payload.items() if v is not None}


def str_to_payload(items):
    return ",".join([f'"{str(item)}"' for item in items])


def convert_tag_list_to_query_language(tags, intersect_tags: bool, prefix: str = ""):
    if tags is None:
        return None

    if not isinstance(tags, list):
        tags = [tags]
    computed_tags = []
    for tag in tags:
        if isinstance(tag, str):
            computed_tags.append(tag)
        else:
            # tag is a Tag, cannot be imported here
            computed_tags.append(tag.name)

    if not computed_tags:
        return None

    if intersect_tags:
        return " and ".join([f'{prefix}tags.name = "{tag}"' for tag in computed_tags])
    else:
        return prefix + "tags.name in (" + str_to_payload(computed_tags) + ")"


def combine_two_ql(q1, q2):
    if q1 is None:
        if q2 is None:
            return None
        else:
            return q2
    else:
        if q2 is None:
            return q1
        else:
            return q1 + " and " + q2


def chunk_list(items: list, chunk_size: int):
    return [(items[i : i + chunk_size], i) for i in range(0, len(items), chunk_size)]


def get_image_shape_with_exif_transpose(image: Image):
    """
        This method reads exif tags of an image and invert width and height if needed.
        Orientation flags that need inversion are : TRANSPOSE, ROTATE_90, TRANSVERSE and ROTATE_270

    Args:
        image: PIL Image to read

    Returns:
        width and height of image
    """
    exif = image.getexif()
    orientation = exif.get(0x0112)

    # Orientation when height and width are inverted :
    # 5: Image.Transpose.TRANSPOSE
    # 6: Image.Transpose.ROTATE_270
    # 7: Image.Transpose.TRANSVERSE
    # 8: Image.Transpose.ROTATE_90

    if orientation in [5, 6, 7, 8]:
        return image.height, image.width
    else:
        return image.width, image.height


def flatten(items: List[List]) -> List:
    return [item for sublist in items for item in sublist]


def flatten_dict(items: List[Dict], filter_value: Callable = lambda _: True) -> Dict:
    final_results = {}
    for res in items:
        for key, value in res.items():
            if key not in final_results:
                final_results[key] = value
            elif filter_value(value):
                final_results[key] += value
    return final_results



import logging
import warnings
from pathlib import Path
from typing import Dict, List, Union

from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from picsellia_annotations.exceptions import FileError, ParsingError
from picsellia_annotations.utils import read_pascal_voc_file
from picsellia_annotations.voc import Object, PascalVOCFile

from picsellia.decorators import exception_handler
from picsellia.exceptions import (
    FileNotFoundException,
    UnparsableAnnotationFileException,
)
from picsellia.sdk.asset import Asset
from picsellia.types.enums import InferenceType

warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger("picsellia")


@exception_handler
@beartype
def parse_file(file_path: Union[Path, str]) -> PascalVOCFile:
    try:
        return read_pascal_voc_file(file_path=file_path)
    except FileError:
        raise FileNotFoundException(f"{file_path} not found")
    except ParsingError as e:
        raise UnparsableAnnotationFileException(
            f"Could not parse VOC file {file_path} because : {e}"
        )


@exception_handler
@beartype
def parse_objects(vocfile: PascalVOCFile) -> List[Object]:
    if isinstance(vocfile.annotation.object, Object):
        return [vocfile.annotation.object]
    else:
        return vocfile.annotation.object


@exception_handler
@beartype
def read_annotations_from_voc_objects(
    inference_type: InferenceType,
    objects: List[Object],
    labels: Dict[str, str],
    asset: Asset,
) -> List[Dict]:
    rectangles: List[Dict] = []
    polygons: List[Dict] = []
    classifications: List[Dict] = []
    for obj in objects:
        label_id = labels[obj.name]

        if inference_type == InferenceType.SEGMENTATION:
            coords = obj.polygon_to_list_coordinates()
            polygons.append({"polygon": coords, "label_id": label_id})
        elif inference_type == InferenceType.OBJECT_DETECTION:
            rectangles.append(
                {
                    "x": int(obj.bndbox.xmin),
                    "y": int(obj.bndbox.ymin),
                    "w": int(obj.bndbox.xmax) - int(obj.bndbox.xmin),
                    "h": int(obj.bndbox.ymax) - int(obj.bndbox.ymin),
                    "label_id": label_id,
                }
            )
        elif inference_type == InferenceType.CLASSIFICATION:
            classifications.append({"label_id": label_id})

    return [
        {
            "asset_id": asset.id,
            "rectangles": rectangles,
            "polygons": polygons,
            "classifications": classifications,
        }
    ]



import logging
import os
from functools import partial
from pathlib import Path
from typing import List, Optional, Union
from uuid import UUID

import orjson
from beartype import beartype
from deprecation import deprecated

import picsellia.exceptions as exceptions
import picsellia.utils as utils
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.datalake import Datalake
from picsellia.sdk.model_context import ModelContext
from picsellia.sdk.model_file import ModelFile
from picsellia.sdk.tag import Tag, TagTarget
from picsellia.sdk.taggable import Taggable
from picsellia.types.enums import Framework, InferenceType, ObjectDataType
from picsellia.types.schemas import ModelVersionSchema

logger = logging.getLogger("picsellia")


class ModelVersion(Dao, Taggable):
    def __init__(self, connexion: Connexion, data: dict):
        Dao.__init__(self, connexion, data)
        Taggable.__init__(self, TagTarget.MODEL_VERSION)

    @property
    def origin_name(self) -> str:
        """Name of origin of this (ModelVersion)"""
        return self._origin_name

    @property
    def origin_id(self) -> UUID:
        """UUID of the origin (Model) of this (ModelVersion)"""
        return self._origin_id

    @property
    def name(self) -> str:
        """Name of this (ModelVersion)"""
        return self._name

    @property
    def version(self) -> int:
        """Version number of this (ModelVersion)"""
        return self._version

    @property
    def type(self) -> InferenceType:
        """Type of this (ModelVersion)"""
        return self._type

    @property
    def framework(self) -> Framework:
        """Framework of this (ModelVersion)"""
        return self._framework

    @property
    def labels(self) -> Optional[dict]:
        return self._labels

    def __str__(self):
        return f"{Colors.BLUE}Version {self.name} of Model '{self.version}' with type {self.type.name} and framework {self.framework.name} {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(foo_dataset.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/model/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/model/version/{self.id}"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/model/version/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> ModelVersionSchema:
        schema = ModelVersionSchema(**data)
        self._origin_name = schema.origin.name
        self._origin_id = schema.origin.id
        self._version = schema.version
        self._name = schema.name
        self._type = schema.type
        self._framework = schema.framework
        self._labels = schema.labels
        return schema

    @exception_handler
    @beartype
    def get_tags(self) -> List[Tag]:
        """Retrieve the tags of your model version.

        Examples:
            ```python
            tags = my_model_version.get_tags()
            assert tags[0].name == "my-model-version-1"
            ```

        Returns:
            List of tags as Tag
        """
        r = self.sync()
        return list(map(partial(Tag, self.connexion), r["tags"]))

    @exception_handler
    @beartype
    def update(
        self,
        labels: Optional[dict] = None,
        docker_image_name: Optional[str] = None,
        docker_flags: Optional[List[str]] = None,
        thumb_object_name: Optional[str] = None,
        notebook_link: Optional[str] = None,
        base_parameters: Optional[dict] = None,
        docker_env_variables: Optional[dict] = None,
        framework: Union[str, Framework, None] = None,
        type: Union[str, InferenceType, None] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        """Update this model version with some new infos.

        Examples:
            ```python
            model_v1.update(docker_image_name="docker.io/model1")
            ```
        """
        if framework:
            framework = Framework.validate(framework)

        if type:
            type = InferenceType.validate(type)

        if type and type not in [
            InferenceType.CLASSIFICATION,
            InferenceType.OBJECT_DETECTION,
            InferenceType.SEGMENTATION,
        ]:
            raise TypeError(f"Type '{type}' not supported yet for model version")

        payload = {
            "labels": labels,
            "docker_image_name": docker_image_name,
            "docker_flags": docker_flags,
            "thumb_object_name": thumb_object_name,
            "notebook_link": notebook_link,
            "base_parameters": base_parameters,
            "docker_env_variables": docker_env_variables,
            "type": type,
            "framework": framework,
            "name": name,
            "description": description,
        }
        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.patch(
            f"/sdk/model/version/{self.id}",
            data=orjson.dumps(filtered_payload),
        ).json()
        self.refresh(r)

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete model version.

        Delete the model in Picsellia database

        Examples:
            ```python
            model_v1.delete()
            ```
        """
        self.connexion.delete(f"/sdk/model/version/{self.id}")
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def get_context(self) -> ModelContext:
        """Get ModelContext of this model

        Examples:
            ```python
            model_v1 = client.get_model(name="my-model").get_version(0)
            context = model_v1.get_context()
            context.get_infos()
            ```

        Returns:
            ModelContext objects that you can use and manipulate
        """
        try:
            r = self.connexion.get(f"/sdk/model/version/{self.id}/context").json()
        except exceptions.ForbiddenError:
            r = self.connexion.get(
                f"/sdk/model/version/{self.id}/public/context"
            ).json()

        return ModelContext(self.connexion, r)

    @exception_handler
    @beartype
    def list_files(self) -> List[ModelFile]:
        """Get a list of ModelFile that were stored with this model

        Examples:
            ```python
            model_v1 = client.get_model(name="my-model").get_version(0)
            files = model_v1.list_files()
            files[0].download()
            ```

        Returns:
            A list of (ModelFile) that you can use and manipulate
        """
        r = self.connexion.get(f"/sdk/model/version/{self.id}/modelfiles").json()
        return list(map(partial(ModelFile, self.connexion), r["items"]))

    @exception_handler
    @beartype
    def get_file(self, name: str) -> ModelFile:
        """Retrieve a ModelFile that were stored with this name into this model

        Examples:
            ```python
            model_v1 = client.get_model(name="my-model").get_version(0)
            file = model_v1.get_file("model-latest")
            file.download()
            ```

        Returns:
            A (ModelFile) that you can use and manipulate
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/sdk/model/version/{self.id}/modelfiles/find", params=params
        ).json()
        return ModelFile(self.connexion, r)

    @exception_handler
    @beartype
    def store(
        self,
        name: str,
        path: Union[str, Path],
        do_zip: bool = False,
        replace: bool = False,
    ) -> ModelFile:
        """Store a file into picsellia storage and attach it to this model.

        Examples:
            ```python
            model.store("model-latest", "./lg_test_file.pb")
            ```
        Arguments:
            name (str): Name of file
            path (str or Path): Path of file to store
            do_zip (bool, optional): If true, zip directory to store it. Defaults to False.
            replace (bool, optional): If true, if a file with given name exists, it will be replaced. Defaults to False.

        Returns:
            A (ModelFile) object
        """
        if not os.path.exists(path):
            raise exceptions.FileNotFoundException(f"{path} not found")

        if do_zip:
            path = utils.zipdir(path)

        filename = os.path.split(path)[-1]
        object_name = self.connexion.generate_model_version_object_name(
            filename, ObjectDataType.MODEL_FILE, self.id
        )
        _, is_large, _ = self.connexion.upload_file(object_name, path)

        # Replacing file
        if replace:
            try:
                model_file = self.get_file(name)
                model_file.delete()
            except exceptions.ResourceNotFoundError:
                logger.debug(f"Model file `{name}` did not exist yet.")

        payload = {
            "name": name,
            "filename": filename,
            "object_name": object_name,
            "large": is_large,
        }
        r = self.connexion.post(
            f"/sdk/model/version/{self.id}/modelfiles",
            data=orjson.dumps(payload),
        ).json()
        model_file = ModelFile(self.connexion, r)
        logger.info(f"{model_file} stored for {self}")
        return model_file

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.7.0",
        details="update_thumbnail method will be removed in the future, you should use the web app instead",
    )
    def update_thumbnail(self, path: Union[str, Path]) -> None:
        """Updates the model thumbnail.

        Update the model thumbnail with given file.
        File size shall be less than 5Mb.

        Examples:
            ```python
            model.update_thumb("test.png")
            ```
        Arguments:
            path (str or Path): Path of the thumbnail you want to push

        Raises:
            FileNotFoundException: If there is no file in given path
            exceptions.InvalidQueryError: If file is too large
            exceptions.PicselliaError: If an unexpected error occurred while uploading file
        """
        if not os.path.isfile(path):
            raise exceptions.FileNotFoundException(f"{path} not found")

        filesize = Path(path).stat().st_size

        if filesize > 5 * 1024 * 1024:
            raise exceptions.BadRequestError("File too large limit is 5Mb")

        filename = os.path.split(path)[-1]
        object_name = self.connexion.generate_model_version_object_name(
            filename, ObjectDataType.MODEL_THUMB, self.id
        )
        self.connexion.upload_file(object_name, path)

        self.update(thumb_object_name=object_name)

    @exception_handler
    @beartype
    def deploy(
        self,
        name: Optional[str] = None,
        target_datalake: Optional[Datalake] = None,
        min_threshold: Optional[float] = None,
    ):
        """Create a (Deployment) for a model.

        This method allows you to create a (Deployment) on Picsellia. You will then have
        access to the monitoring dashboard and the model management part!

        Examples:
            ```python
            model_version = client.get_model(name="my-awesome-model").get_version(0)
            deployment = model_version.deploy(name="my-awesome-deployment", min_threshold=0.5)
            ```

        Arguments:
            name (str): Name of your deployment. Defaults to a random name.
            min_threshold (float): Threshold of detection scores used by models when predicting. Defaults to 0.
            target_datalake (Datalake): Datalake to use when data are pushed into Picsellia.
                Defaults to organization default Datalake.


        Returns:
            A (Deployment)
        """
        from picsellia.sdk.deployment import Deployment

        payload = {
            "name": name,
            "min_threshold": min_threshold,
        }
        if target_datalake is not None:
            payload["target_datalake_id"] = target_datalake.id
        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.post(
            f"/sdk/model/version/{self.id}/deploy",
            data=orjson.dumps(filtered_payload),
        ).json()
        deployment = Deployment(self.connexion, r)
        logger.info(f"{deployment} created")
        return deployment



import logging
import subprocess
import warnings
from functools import partial
from typing import Dict, List, Union
from uuid import UUID

import orjson
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import BadConfigurationScanError
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.experiment import Experiment
from picsellia.sdk.scan_file import ScanFile
from picsellia.types.enums import RunStatus
from picsellia.types.schemas import RunSchema

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Run(Dao):
    def __init__(self, connexion: Connexion, data: dict) -> None:
        Dao.__init__(self, connexion, data)

    @property
    def scan_id(self) -> UUID:
        """UUID of the (Scan) holding this (Run)"""
        return self._scan_id

    @property
    def experiment_id(self) -> UUID:
        """UUID of the (Experiment) of this (Run)"""
        return self._experiment_id

    @property
    def order(self) -> int:
        """Order of this (Run)"""
        return self._order

    @property
    def parameters(self) -> Dict:
        """Parameters of this (Run)"""
        return self._parameters

    @property
    def status(self) -> RunStatus:
        """Status of this (Run)"""
        return self._status

    def __str__(self):
        return f"{Colors.BLUE}Run {self.order}{Colors.ENDC} of scan {self.scan_id} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/run/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> RunSchema:
        schema = RunSchema(**data)
        self._order = schema.order
        self._parameters = schema.parameters
        self._status = schema.status
        self._scan_id = schema.scan_id
        self._experiment_id = schema.experiment_id
        return schema

    @exception_handler
    @beartype
    def update(self, status: Union[RunStatus, str]) -> None:
        """Update this run.

        Examples:
            ```python
            run.update(status=RunStatus.TERMINATED)
            ```
        """
        payload = {"status": RunStatus.validate(status)}
        r = self.connexion.patch(
            f"/sdk/run/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def end(self) -> None:
        """End a run

        Examples:
            ```python
            run.end()
            ```
        """
        r = self.connexion.post(f"/sdk/run/{self.id}/end").json()
        self.refresh(r["run"])
        logger.info(f"{self} ended: {r['message']}")

    @exception_handler
    @beartype
    def get_script(
        self,
    ) -> ScanFile:
        """Retrieve the script of this run.

        Returns:
            A (ScanFile) object
        """
        r = self.connexion.get(f"/sdk/scan/{self.scan_id}").json()

        if "script_id" not in r or r["script_id"] is None:
            raise BadConfigurationScanError("This scan has no script.")

        r = self.connexion.get(f"/sdk/scan/file/{r['script_id']}").json()
        return ScanFile(self.connexion, r)

    @exception_handler
    @beartype
    def list_data_files(
        self,
    ) -> List[ScanFile]:
        """List all data files of this run

        Examples:
            ```python
            files = run.list_data_files()
            ```

        Returns:
            List of (ScanFile) object
        """
        r = self.connexion.get(f"/sdk/scan/{self.scan_id}/scanfiles").json()
        return list(map(partial(ScanFile, self.connexion), r["items"]))

    @exception_handler
    @beartype
    def install_requirements(self) -> None:
        """Install requirements from the run requirements dictionary.

        Examples:
            ```python
            run.install_requirements()
            ```
        """
        r = self.connexion.get(f"/sdk/scan/{self.scan_id}").json()
        requirements = r["requirements"]

        for module in requirements:
            name = (
                f"{module['package']}=={module['version']}"
                if module["version"] != ""
                else module["package"]
            )
            subprocess.call(["pip", "install", name])

    @exception_handler
    @beartype
    def get_experiment(self) -> Experiment:
        """Retrieve linked experiment

        Examples:
            ```python
            my_experiment = run.get_experiment()
            ```

        Returns:
            An (Experiment) object linked to this run
        """
        r = self.connexion.get(f"/sdk/experiment/{self.experiment_id}").json()
        return Experiment(self.connexion, r)

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this run from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            run.delete()
            ```
        """
        self.connexion.delete(f"/sdk/run/{self.id}")
        logger.info(f"{self} deleted.")



import logging
import warnings

import orjson
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.types.schemas import DataSourceSchema

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class DataSource(Dao):
    def __init__(self, connexion: Connexion, data: dict):
        Dao.__init__(self, connexion, data)

    def __str__(self):
        return f"{Colors.GREEN}Data source '{self.name}'{Colors.ENDC} (id: {self.id})"

    @property
    def name(self) -> str:
        """Name of this (DataSource)"""
        return self._name

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> DataSourceSchema:
        schema = DataSourceSchema(**data)
        self._name = schema.name
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/data/source/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def update(self, name: str) -> None:
        """Update this data source with a new name.

        Examples:
            ```python
            sdk_source.update(name="new name")
            ```
        """
        payload = {"name": name}
        r = self.connexion.patch(
            f"/sdk/data/source/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this data source from the platform.
        All data with this source will not have source anymore

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            sdk_source.delete()
            ```
        """
        self.connexion.delete(f"/sdk/data/source/{self.id}")
        logger.info(f"{self} deleted.")



import warnings

from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.types.schemas import WorkerSchema

warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Worker(Dao):
    def __init__(self, connexion: Connexion, data: dict):
        Dao.__init__(self, connexion, data)

    @property
    def username(self) -> str:
        """Username of the Worker"""
        return self._username

    def __str__(self):
        return f"{Colors.UNDERLINE}Worker '{self.username}' {Colors.ENDC}"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/worker/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> WorkerSchema:
        schema = WorkerSchema(**data)
        self._username = schema.collaborator.username
        return schema

    @exception_handler
    @beartype
    def get_infos(self) -> dict:
        """Retrieve worker info

        Examples:
            ```python
            worker = my_dataset.list_workers()[0]
            print(worker.get_infos())
            ```

        Returns:
            A dict with data of the worker
        """
        return {"username": self.username}



import logging
import mimetypes
import os
import warnings
from io import BufferedReader
from pathlib import Path
from typing import IO, Any, BinaryIO, Dict, Iterable, Mapping, Optional, Tuple, Union
from uuid import UUID

import orjson
import requests
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from picsellia_connexion_services import TokenServiceConnexion
from requests.exceptions import ConnectionError

import picsellia.exceptions as exceptions
from picsellia import __version__
from picsellia.decorators import exception_handler, retry
from picsellia.types.enums import ObjectDataType
from picsellia.utils import (
    handle_response,
    print_line_return,
    print_start_section,
    print_stop_section,
)

warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)

logger = logging.getLogger("picsellia")


class Connexion(TokenServiceConnexion):
    def __init__(
        self, host: str, api_token: str, content_type: str = "application/json"
    ) -> None:
        super().__init__(
            host, api_token, authorization_key="Bearer", content_type=content_type
        )
        self._connector_id = None
        self._organization_id = None
        self.add_header("User-Agent", f"Picsellia-SDK/{__version__}")

    @property
    def connector_id(self):
        if self._connector_id is None:
            raise exceptions.NoConnectorFound(
                "This organization has no default connector, and connect retrieve and upload files."
            )
        return self._connector_id

    @connector_id.setter
    def connector_id(self, value):
        self._connector_id = value

    @property
    def organization_id(self):
        return self._organization_id

    @organization_id.setter
    def organization_id(self, value):
        self._organization_id = value
        self.add_header("X-Picsellia-Organization", str(self._organization_id))

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, Connexion):
            return self.host == __o.host and self.api_token == __o.api_token

        return False

    @handle_response
    def get(self, path: str, params: Optional[dict] = None, stream=False):
        return super().get(path=path, params=params, stream=stream)

    @handle_response
    def xget(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[Tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
        stream=False,
    ):
        return super().xget(path=path, data=data, params=params, stream=stream)

    @handle_response
    def post(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[Tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
        files: Optional[Any] = None,
    ):
        return super().post(path=path, data=data, params=params, files=files)

    @handle_response
    def put(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[Tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
    ):
        return super().put(path=path, data=data, params=params)

    @handle_response
    def patch(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[Tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
    ):
        return super().patch(path=path, data=data, params=params)

    @handle_response
    def delete(
        self,
        path: str,
        data: Union[
            None, str, bytes, Mapping[str, Any], Iterable[Tuple[str, str, None]], IO
        ] = None,
        params: Optional[dict] = None,
    ):
        return super().delete(path=path, data=data, params=params)

    ##############################################################
    # ------------------------- UPLOAD ------------------------- #
    ##############################################################
    @exception_handler
    @beartype
    def _generate_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        connector_id: Optional[UUID] = None,
        context: Optional[Dict[str, UUID]] = None,
    ) -> str:
        if connector_id is None:
            connector_id = self.connector_id

        payload = {"filename": filename, "type": object_name_type}
        if context:
            payload["context"] = context

        r = self.post(
            path=f"/sdk/organization/{self.organization_id}/connector/{connector_id}/generate_object_name",
            data=orjson.dumps(payload),
        ).json()
        return r["object_name"]

    @beartype
    def generate_data_object_name(
        self, filename: str, connector_id: Optional[UUID] = None
    ):
        return self._generate_object_name(filename, ObjectDataType.DATA, connector_id)

    @beartype
    def generate_job_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        job_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in [ObjectDataType.LOGGING]:
            raise RuntimeError(
                f"Cannot generate job object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"job_id": job_id},
        )

    @beartype
    def generate_experiment_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        experiment_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in (
            ObjectDataType.ARTIFACT,
            ObjectDataType.LOG_IMAGE,
            ObjectDataType.LOGGING,
        ):
            raise RuntimeError(
                f"Cannot generate experiment object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"experiment_id": experiment_id},
        )

    @beartype
    def generate_model_version_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        model_version_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type not in (
            ObjectDataType.MODEL_THUMB,
            ObjectDataType.MODEL_FILE,
        ):
            raise RuntimeError(
                f"Cannot generate model version object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"model_version_id": model_version_id},
        )

    @beartype
    def generate_project_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        project_id: UUID,
        connector_id: Optional[UUID] = None,
    ):
        if object_name_type != ObjectDataType.SCAN_FILE:
            raise RuntimeError(
                f"Cannot generate project object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"project_id": project_id},
        )

    @exception_handler
    @beartype
    def _init_upload(self, path: Union[str, Path]) -> Tuple[bool, str]:
        if not os.path.isfile(path):
            raise exceptions.FileNotFoundException(f"{path} not found")

        is_large = Path(path).stat().st_size > 5 * 1024 * 1024
        content_type = mimetypes.guess_type(path, strict=False)[0]
        if content_type is None:
            content_type = "text/plain"

        return is_large, content_type

    @exception_handler
    @beartype
    def _generate_presigned_url(
        self, object_name: str, content_type: str, connector_id: UUID
    ):
        payload = {"object_name": object_name, "content_type": content_type}
        r = self.post(
            path=f"/sdk/connector/{connector_id}/generate_presigned_url",
            data=orjson.dumps(payload),
        )

        if r.status_code != 200:
            raise exceptions.DistantStorageError("Errors while getting a presigned url")

        r = r.json()["presigned_url_data"]
        if "url" not in r:
            raise exceptions.DistantStorageError(
                "Errors while getting a presigned url. Unparsable response"
            )

        return r

    @retry((KeyError, ValueError, exceptions.PicselliaError), total_tries=3)
    @exception_handler
    @beartype
    def _do_upload_file(
        self, object_name: str, path: Union[str, Path], presigned_url_data: dict
    ) -> requests.Response:
        if not os.path.isfile(path):
            raise exceptions.FileNotFoundException(f"{path} not found")

        with open(path, "rb") as file:
            try:
                response = requests.post(
                    url=presigned_url_data["url"],
                    data=presigned_url_data["fields"],
                    files={"file": (object_name, file)},
                )
            except Exception:
                raise exceptions.NetworkError("Impossible to upload file")

        if response.status_code != 204:
            logging.error(
                f"Could not upload {path} into S3. Status code: {response.status_code}"
            )
            raise exceptions.DistantStorageError(f"Could not upload {path} into S3")

        return response

    @exception_handler
    @beartype
    def upload_file(
        self,
        object_name: str,
        path: Union[str, Path],
        connector_id: Optional[UUID] = None,
    ) -> Tuple[requests.Response, bool, str]:
        """Upload a single file to the server.
        If file is bigger than 5Mb, it will send it by multipart.

        Arguments:
            path (str): [Absolute path to the file]
            object_name (str): [Bucket prefix s3]
            connector_id (UUID): Use this custom connector id in some case
        """
        if connector_id is None:
            connector_id = self.connector_id

        is_large, content_type = self._init_upload(path)

        if is_large:
            try:
                return (
                    self._upload_large_file(object_name, path, connector_id),
                    is_large,
                    content_type,
                )
            except exceptions.BadRequestError:
                logger.warning(
                    "This file is large but it is impossible to use multipart for this upload."
                    "Trying to upload in only one upload."
                )

        presigned_url_data = self._generate_presigned_url(
            object_name, content_type, connector_id
        )

        response: requests.Response = self._do_upload_file(
            object_name, path, presigned_url_data
        )

        if response.status_code >= 300:
            raise exceptions.UploadFailed(
                "Something wrong happened while storing this file"
            )

        return response, is_large, content_type

    @exception_handler
    @beartype
    def _init_multipart(self, object_name: str, connector_id: UUID) -> str:
        """Initialize a multipart push

        Arguments:
            object_name (str): object name to download

        Returns:
            The upload id if everything went well
        """

        payload = {"object_name": object_name}

        r = self.post(
            path=f"/sdk/connector/{connector_id}/init_multipart_upload",
            data=orjson.dumps(payload),
        )

        r = r.json()
        if "upload_id" not in r:
            raise exceptions.DistantStorageError(
                "Response when initiating multipart is unparsable"
            )

        return r["upload_id"]

    @exception_handler
    @beartype
    def _get_url_for_part(
        self, object_name: str, upload_id: str, no_part: int, connector_id: UUID
    ) -> str:
        """Get a pre-signed url to upload a part of a large file"""
        payload = {
            "object_name": object_name,
            "upload_id": upload_id,
            "part_no": no_part,
        }
        r = self.post(
            path=f"/sdk/connector/{connector_id}/generate_part_presigned_url",
            data=orjson.dumps(payload),
        )

        r = r.json()
        if "url" not in r:
            raise exceptions.NetworkError(
                "Response when getting an url for a part is unparsable"
            )

        return r["url"]

    @exception_handler
    @beartype
    def _upload_parts(
        self,
        object_name: str,
        upload_id: str,
        path: Union[str, Path],
        connector_id: UUID,
    ) -> list:
        if not os.path.exists(path):
            raise exceptions.FileNotFoundException(
                "Impossible to upload part in an empty filepath"
            )

        max_size = 5 * 1024 * 1024
        file_size = os.path.getsize(path)
        upload_by = int(file_size / max_size) + 1
        print_start_section()
        with open(path, "rb") as file:
            parts = []
            for part in range(1, upload_by + 1):
                etag = self._do_upload_one_part(
                    object_name, upload_id, part, connector_id, file, max_size
                )
                parts.append({"ETag": etag, "PartNumber": part})
        print_stop_section()
        print_line_return()
        return parts

    @exception_handler
    @beartype
    def _do_upload_one_part(
        self,
        object_name: str,
        upload_id: str,
        part: int,
        connector_id: UUID,
        # Ensures backport with beartype and python
        file: Union[BufferedReader, BinaryIO],
        max_size: int,
    ):
        try:
            url = self._get_url_for_part(object_name, upload_id, part, connector_id)
            file_data = file.read(max_size)
            res = requests.put(url, data=file_data)
        except Exception:
            raise exceptions.DistantStorageError("Impossible to get an url for part")

        if res.status_code != 200:
            raise exceptions.DistantStorageError(
                f"Impossible to put part no {part}\n because {res.text}"
            )
        return res.headers["ETag"]

    @exception_handler
    @beartype
    def _complete_part_upload(
        self, object_name: str, upload_id: str, parts: list, connector_id: UUID
    ) -> requests.Response:
        """
        Complete the upload a part of a large file
        """
        payload = {
            "object_name": object_name,
            "upload_id": upload_id,
            "parts": parts,
        }
        try:
            r = self.post(
                path=f"/sdk/connector/{connector_id}/complete_part_upload",
                data=orjson.dumps(payload),
            )
        except Exception as e:
            raise exceptions.NetworkError(f"Impossible to complete part url : {e}")

        return r

    @exception_handler
    @beartype
    def _upload_large_file(
        self, object_name: str, path: Union[str, Path], connector_id: UUID
    ) -> requests.Response:
        """Upload a single large file to the server. It will be sent by multipart.

        Arguments:
            object_name (str): [Bucket prefix s3]
            path (str): [Absolute path to the file]
            connector_id (UUID): [Connector id in Platform to use. Defaults to organization default connector]
        """
        if connector_id is None:
            connector_id = self.connector_id

        upload_id: str = self._init_multipart(object_name, connector_id)
        parts: list = self._upload_parts(object_name, upload_id, path, connector_id)
        response: requests.Response = self._complete_part_upload(
            object_name, upload_id, parts, connector_id
        )

        if response.status_code >= 300:
            raise exceptions.UploadFailed(
                "Something wrong happened while storing this file"
            )

        return response

    ##############################################################
    # ------------------------ DOWNLOAD ------------------------ #
    ##############################################################
    @exception_handler
    @beartype
    def init_download(
        self, object_name: str, connector_id: Optional[UUID] = None
    ) -> str:
        """Retrieve a presigned url of this object name in order to download it"""
        if connector_id is None:
            connector_id = self.connector_id

        payload = {"object_name": object_name}
        r = self.post(
            path=f"/sdk/connector/{connector_id}/retrieve_presigned_url",
            data=orjson.dumps(payload),
        )

        if r.status_code != 200:
            raise exceptions.DistantStorageError("Errors while getting a presigned url")

        r = r.json()
        if "presigned_url" not in r:
            raise exceptions.DistantStorageError(
                "Errors while getting a presigned url. Unparsable response"
            )

        return r["presigned_url"]

    @exception_handler
    @beartype
    def do_download_file(
        self,
        path: Union[str, Path],
        url: str,
        is_large: bool,
        force_replace: bool,
        retry_count: int = 1,
    ) -> bool:
        try:
            return self._do_download_file(path, url, is_large, force_replace)
        except (exceptions.NetworkError, ConnectionError) as e:
            if retry_count <= 0:
                raise exceptions.DownloadError(
                    f"Could not download {url} into {path}"
                ) from e
            logger.error(
                f"Could not download because of a NetworkError. Retrying to download {path}"
            )
            return self.do_download_file(
                path,
                url,
                is_large,
                force_replace,
                retry_count=retry_count - 1,
            )

    @exception_handler
    @beartype
    def _do_download_file(
        self,
        path: Union[str, Path],
        url: str,
        is_large: bool,
        force_replace: bool,
    ) -> bool:
        """Retrieve a presigned url of this object name in order to download it"""
        if os.path.exists(path) and not force_replace:
            return False

        parent_path = Path(path).parent.absolute()
        if not os.path.exists(parent_path):
            os.makedirs(parent_path)

        response = requests.get(url, stream=is_large)

        if response.status_code == 429 or (500 <= response.status_code < 600):
            raise exceptions.NetworkError(
                f"Response status code is {response.status_code}. Could not get {url}"
            )

        response.raise_for_status()

        total_length = response.headers.get("content-length")
        if total_length is None:
            raise exceptions.NetworkError(
                "Downloaded content is empty but response is 200"
            )

        with open(path, "wb") as handler:
            if not is_large:
                handler.write(response.content)
            else:
                for data in response.iter_content(chunk_size=4096):
                    handler.write(data)

        return True



import base64
import logging
import mimetypes
import warnings
from functools import partial
from pathlib import Path
from typing import List, Optional, Union

import orjson
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from deprecation import deprecated
from picsellia_connexion_services import JwtServiceConnexion

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import (
    BadConfigurationContinuousTrainingError,
    BadRequestError,
    ContentTypeUnknown,
    MonitorError,
    NoShadowModel,
    PicselliaError,
    PredictionError,
)
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.datalake import Datalake
from picsellia.sdk.dataset import DatasetVersion
from picsellia.sdk.datasource import DataSource
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.project import Project
from picsellia.sdk.tag import Tag
from picsellia.sdk.taggable import Taggable
from picsellia.types.enums import (
    ContinuousDeploymentPolicy,
    ContinuousTrainingTrigger,
    ContinuousTrainingType,
    InferenceType,
    ServiceMetrics,
    SupportedContentType,
    TagTarget,
)
from picsellia.types.schemas import DeploymentSchema
from picsellia.types.schemas_prediction import PredictionFormat

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Deployment(Dao, Taggable):
    def __init__(self, connexion: Connexion, data: dict):
        Dao.__init__(self, connexion, data)
        Taggable.__init__(self, TagTarget.DEPLOYMENT)

        deployment = self.refresh(data)
        if deployment.oracle_host is not None:
            try:
                self._oracle_connexion = JwtServiceConnexion(
                    deployment.oracle_host,
                    {
                        "api_token": self.connexion.api_token,
                        "deployment_id": str(self.id),
                    },
                    login_path="/api/auth/login",
                )
                if self._oracle_connexion.jwt is None:  # pragma: no cover
                    raise PicselliaError("Cannot authenticate to oracle")

                logging.info(
                    f"Connected with monitoring service at {deployment.oracle_host}"
                )
            except Exception as e:
                logger.error(
                    f"Could not bind {self} with our monitoring service at {deployment.oracle_host} because : {e}"
                )
                self._oracle_connexion.session.close()
                self._oracle_connexion = None
        else:  # pragma: no cover
            self._oracle_connexion = None

        if deployment.serving_host is not None:
            try:
                self._serving_connexion = JwtServiceConnexion(
                    deployment.serving_host,
                    {
                        "api_token": self.connexion.api_token,
                        "deployment_id": str(self.id),
                    },
                    login_path="/api/login",
                )
                if self._serving_connexion.jwt is None:  # pragma: no cover
                    raise PicselliaError("Cannot authenticate to serving")
                logging.info(
                    f"Connected with serving service at {deployment.serving_host}"
                )
            except Exception as e:
                logger.error(
                    f"Could not bind {self} with our serving service at {deployment.serving_host} because : {e}"
                )
                self._serving_connexion.session.close()
                self._serving_connexion = None
        else:  # pragma: no cover
            self._serving_connexion = None

    @property
    def name(self) -> str:
        """Name of this (Deployment)"""
        return self._name

    @property
    def type(self) -> InferenceType:
        """Type of this (Deployment)"""
        return self._type

    @property
    def oracle_connexion(self) -> JwtServiceConnexion:
        assert (
            self._oracle_connexion is not None
        ), "You can't use this function with this deployment. Please contact the support."
        return self._oracle_connexion

    @property
    def serving_connexion(self) -> JwtServiceConnexion:
        assert (
            self._serving_connexion is not None
        ), "You can't use this function with this deployment. Please contact the support."
        return self._serving_connexion

    def __str__(self):
        return f"{Colors.CYAN}Deployment '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> DeploymentSchema:
        schema = DeploymentSchema(**data)
        self._name = schema.name
        self._type = schema.type
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/deployment/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def get_tags(self) -> List[Tag]:
        """Retrieve the tags of your deployment.

        Examples:
            ```python
            tags = deployment.get_tags()
            assert tags[0].name == "cool"
            ```

        Returns:
            List of tags as (Tag)
        """
        r = self.sync()
        return list(map(partial(Tag, self.connexion), r["tags"]))

    @exception_handler
    @beartype
    def retrieve_information(self) -> dict:
        """Retrieve some information about this deployment from service.

        Examples:
            ```python
            my_deployment.retrieve_information()
            ```
        """
        return self.oracle_connexion.get(path=f"/api/deployment/{self.id}").json()

    @exception_handler
    @beartype
    def update(
        self,
        name: Optional[str] = None,
        target_datalake: Optional[Datalake] = None,
        min_threshold: Optional[float] = None,
    ) -> None:
        """Update this deployment with a new name, another target datalake or a minimum threshold

        Examples:
            ```python
            a_tag.update(name="new name", min_threshold=0.4)
            ```
        Args:
            name (str, optional): New name of the deployment
            target_datalake (Datalake, optional): Datalake where data will be uploaded on new prediction
            min_threshold (float, optional): Minimum confidence threshold.
                    Serving will filter detection boxes or masks that have a detection score lower than this threshold

        """
        payload = {}
        if name is not None:
            payload["name"] = name

        if min_threshold is not None:
            payload["min_threshold"] = min_threshold

        if target_datalake is not None:
            payload["target_datalake_id"] = target_datalake.id

        r = self.connexion.patch(
            f"/sdk/deployment/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self, force_delete: bool = False) -> None:
        self.connexion.delete(
            f"/sdk/deployment/{self.id}", params={"force_delete": force_delete}
        )
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def set_model(self, model_version: ModelVersion) -> None:
        """Update this deployment with a new name.

        Examples:
            ```python
            a_tag.update(name="new name")
            ```
        """
        payload = {"model_version_id": model_version.id}

        self.connexion.post(
            f"/sdk/deployment/{self.id}/model", data=orjson.dumps(payload)
        ).json()
        logger.info(f"{self} model is now {model_version}")

    @exception_handler
    @beartype
    def get_model_version(self) -> ModelVersion:
        """Retrieve currently used model version

        Examples:
            ```python
            model_version = deployment.get_model()
            ```

        Returns:
            A (Model) object
        """
        r = self.sync()

        r = self.connexion.get(f"/sdk/model/version/{r['model_version_id']}").json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def set_shadow_model(self, shadow_model_version: ModelVersion) -> None:
        """Update this deployment with a new name.

        Examples:
            ```python
            a_tag.update(name="new name")
            ```
        """
        payload = {"model_version_id": shadow_model_version.id}

        self.connexion.post(
            f"/sdk/deployment/{self.id}/shadow", data=orjson.dumps(payload)
        ).json()
        logger.info(f"{self} shadow model is now {shadow_model_version}")

    @exception_handler
    @beartype
    def get_shadow_model(self) -> ModelVersion:
        """Retrieve currently used shadow model

        Examples:
            ```python
            shadow_model = deployment.get_shadow_model()
            ```

        Returns:
            A (Model) object
        """
        r = self.sync()
        if "shadow_model_version_id" not in r or r["shadow_model_version_id"] is None:
            raise NoShadowModel("This deployment has no shadow model")

        r = self.connexion.get(
            f"/sdk/model/version/{r['shadow_model_version_id']}"
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def predict(
        self,
        file_path: Union[str, Path],
        tags: Union[str, Tag, List[Union[Tag, str]], None] = None,
        source: Union[str, DataSource, None] = None,
    ) -> dict:
        """Run a prediction on our Serving platform

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.predict('image_420.png', tags=["gonna", "give"], source="camera-1")
            ```
        Arguments:
            tags (str, (Tag), list of str or Tag, optional): a list of tag to add to the data that will be created on the platform
            source (str or DataSource, optional): a source to attach to the data that will be created on the platform.
            file_path (str or Path): path to the image to predict

        Returns:
            A (dict) with information of the prediction
        """

        sent_tags = []
        if tags:
            if isinstance(tags, str) or isinstance(tags, Tag):
                tags = [tags]

            for tag in tags:
                if isinstance(tag, Tag):
                    sent_tags.append(tag.name)
                else:
                    sent_tags.append(tag)

        if isinstance(source, DataSource):
            source = source.name

        with open(file_path, "rb") as file:
            files = {"media": file}

            payload = {"tags": sent_tags}
            if source:
                payload["source"] = source

            resp = self.serving_connexion.post(
                path=f"/api/deployment/{self.id}/predict",
                data=payload,
                files=files,
            )

            if resp.status_code != 200:  # pragma: no cover
                raise PredictionError(f"Could not predict because {resp.text}")

            return resp.json()

    @exception_handler
    @beartype
    def setup_feedback_loop(self, dataset_version: DatasetVersion) -> None:
        """Set up the Feedback Loop for a Deployment.
        This way, you will be able to push reviewed predictions to given Dataset.
        This is a great option to increase your training set with quality data.

        Examples:
            ```python
            dataset_version = client.get_dataset("my-dataset").get_version("latest")
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.setup_feedback_loop(dataset_version)
            ```
        Arguments:
            dataset_version (DatasetVersion): a connected (DatasetVersion)
        """
        payload = {
            "dataset_version_id": dataset_version.id,
        }
        self.connexion.post(
            f"/sdk/deployment/{self.id}/pipeline/fl/setup",
            data=orjson.dumps(payload),
        )
        logger.info(
            f"Feedback loop set for {self}, now you will be able to add predictions to {dataset_version}"
        )

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.6.0",
        details="check_feedback_loop_status method will be removed in the future",
    )
    def check_feedback_loop_status(self) -> None:
        """Refresh feedback loop status of this deployment.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.check_feedback_loop_status()
            ```
        """
        r = self.connexion.get(f"/sdk/deployment/{self.id}/pipeline/fl/check").json()
        feedback_loop_status = r["feedback_loop_status"]
        logger.info(f"Feedback loop status is {feedback_loop_status}")

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.6.0",
        details="disable_feedback_loop method will be removed in the future",
    )
    def disable_feedback_loop(self) -> None:
        """Disable the Feedback Loop for a Deployment.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.disable_feedback_loop()
            ```
        """
        self.connexion.put(f"/sdk/deployment/{self.id}/pipeline/fl/disable")
        logger.info(f"Feedback loop for {self} is disabled.")

    @exception_handler
    @beartype
    def toggle_feedback_loop(self, active: bool) -> None:
        """Toggle feedback loop for this deployment

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.toggle_feedback_loop(
                dataset
            )
            ```
        Arguments:
            active (bool): (des)activate feedback loop
        """
        payload = {"active": active}
        self.connexion.put(
            f"/sdk/deployment/{self.id}/pipeline/fl",
            data=orjson.dumps(payload),
        )
        logger.info(
            f"Feedback loop for {self} is now {'active' if active else 'deactivated'}"
        )

    @exception_handler
    @beartype
    def set_training_data(self, dataset_version: DatasetVersion) -> None:
        """This will give the training data reference to the deployment,
         so we can compute metrics based on this training data distribution in our Monitoring service

        Examples:
            ```python
            dataset_version = client.get_dataset("my-dataset").get_version("latest")
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.set_training_data(dataset_version)
            ```
        Arguments:
            dataset_version (DatasetVersion): a connected (DatasetVersion)
        """
        payload = {
            "dataset_version_id": dataset_version.id,
        }
        self.connexion.post(
            f"/sdk/deployment/{self.id}/pipeline/td/setup",
            data=orjson.dumps(payload),
        )
        logger.info(f"Training Data set for {self} from {dataset_version}")

    @exception_handler
    @beartype
    def check_training_data_metrics_status(self) -> str:
        """Refresh the status of the metrics compute over the training data distribution.
        Set up can take some time, so you can check current state with this method.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.check_training_data_metrics_status()
            ```
        """
        r = self.connexion.get(f"/sdk/deployment/{self.id}/pipeline/td/check").json()
        training_data_status = r["training_data_status"]
        logger.info(f"Training Data status is {training_data_status}")
        return training_data_status

    @exception_handler
    @beartype
    def disable_training_data_reference(self) -> None:
        """Disable the reference to the training data in this Deployment.
        This means that you will not be able to see supervised metrics from the dashboard anymore.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.disable_training_data_reference()
            ```
        """
        self.connexion.put(f"/sdk/deployment/{self.id}/pipeline/td/disable")
        logger.info(f"Training Data for {self} is disabled.")

    @exception_handler
    @beartype
    def setup_continuous_training(
        self,
        project: Project,
        dataset_version: DatasetVersion,
        model_version: ModelVersion,
        trigger: Union[str, ContinuousTrainingTrigger] = None,
        threshold: Optional[int] = None,
        experiment_parameters: Optional[dict] = None,
        scan_config: Optional[dict] = None,
    ) -> None:
        """Initialize and activate the continuous training features of picsellia. 🥑
           A Training will be triggered using the configured Dataset
           and Model as base whenever your Deployment pipeline hit the trigger.

            There is 2 types of continuous training different:
            You can launch a continuous training via Scan configuration or via Experiment
            You need to give whether `experiment_parameters` or `scan_config` but not both
            For scan configuration: [more info](https://doc.picsellia.com/docs/initialize-a-scan).

        Examples:
            We want to set up a continuous training pipeline that will be trigger
            every 150 new predictions reviewed by your team.
            We will use the same training parameters as those used when building the first model.

            ```python
            deployment = client.get_deployment("awesome-deploy")
            project = client.get_project(name="my-project")
            dataset_version = project.get_dataset(name="my-dataset").get_version("latest")
            model_version = client.get_model(name="my-model").get_version(0)
            experiment = model_version.get_source_experiment()
            parameters = experiment.get_log('parameters')
            feedback_loop_trigger = 150
            deployment.setup_continuous_training(
                project, dataset_version, model_version,
                threshold=150, experiment_parameters=experiment_parameters
            )
            ```
        Arguments:
            project (Project): The project that will host your pipeline.
            dataset_version (DatasetVersion): The Dataset that will be used as training data for your training.
            model_version (ModelVersion):  The exported Model to perform transfer learning from.
            threshold (int): Number of images that need to be review to trigger the training.
            trigger (ContinuousTrainingTrigger): Type of trigger to use when there is enough reviews.
            experiment_parameters (Optional[dict], optional):  Training parameters. Defaults to None.
            scan_config (Optional[dict], optional): Scan configuration dict. Defaults to None.
        """
        payload = {
            "project_id": project.id,
            "dataset_version_id": dataset_version.id,
            "model_version_id": model_version.id,
        }

        if trigger is not None and threshold is not None:
            payload["trigger"] = ContinuousTrainingTrigger.validate(trigger)
            payload["threshold"] = threshold

        if experiment_parameters is not None:
            if scan_config is not None:
                raise BadConfigurationContinuousTrainingError(
                    "You cannot give both experiment_parameters and scan_config"
                )
            else:
                payload["training_type"] = ContinuousTrainingType.EXPERIMENT
                payload["experiment_parameters"] = experiment_parameters
        else:
            if scan_config is not None:
                payload["training_type"] = ContinuousTrainingType.SCAN
                payload["scan_config"] = scan_config
            else:
                raise BadConfigurationContinuousTrainingError(
                    "You need to give experiment_parameters or scan_config"
                )

        self.connexion.post(
            f"/sdk/deployment/{self.id}/pipeline/ct",
            data=orjson.dumps(payload),
        )
        logger.info(f"Continuous training setup for {self}\n")

    @exception_handler
    @beartype
    def toggle_continuous_training(self, active: bool) -> None:
        """Update your continuous training pipeline.

        Examples:
            ```python
            deployment = client.get_deployment("awesome-deploy")
            deployment.update_continuous_training(active=False)
            ```
        """
        payload = {"active": active}
        self.connexion.put(
            f"/sdk/deployment/{self.id}/pipeline/ct",
            data=orjson.dumps(payload),
        )
        logger.info(
            f"Continuous training for {self} is now {'active' if active else 'deactivated'}"
        )

    @exception_handler
    @beartype
    def setup_continuous_deployment(
        self, policy: Union[ContinuousDeploymentPolicy, str]
    ) -> None:
        """Set up the continuous deployment for this pipeline

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.setup_continuous_deployment(ContinuousDeploymentPolicy.DEPLOY_MANUAL)
            ```
        Arguments:
            policy (ContinuousDeploymentPolicy): policy to use
        """
        payload = {"policy": ContinuousDeploymentPolicy.validate(policy)}
        self.connexion.post(
            f"/sdk/deployment/{self.id}/pipeline/cd",
            data=orjson.dumps(payload),
        )
        logger.info(f"Continuous deployment setup for {self} with policy {policy}\n")

    @exception_handler
    @beartype
    def toggle_continuous_deployment(self, active: bool) -> None:
        """Toggle continuous deployment for this deployment

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.toggle_continuous_deployment(
                dataset
            )
            ```
        Arguments:
            active (bool): (des)activate continuous deployment
        """
        payload = {"active": active}
        self.connexion.put(
            f"/sdk/deployment/{self.id}/pipeline/cd",
            data=orjson.dumps(payload),
        )
        logger.info(
            f"Continuous deployment for {self} is now {'active' if active else 'deactivated'}"
        )

    @exception_handler
    @beartype
    def get_stats(
        self,
        service: ServiceMetrics,
        model_version: Optional[ModelVersion] = None,
        from_timestamp: Optional[float] = None,
        to_timestamp: Optional[float] = None,
        since: Optional[int] = None,
        includes: Optional[List[str]] = None,
        excludes: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
    ) -> dict:
        """Retrieve stats of this deployment stored in Picsellia environment.

        Mandatory param is "service" an enum of type ServiceMetrics. Values possibles are :
            PREDICTIONS_OUTLYING_SCORE
            PREDICTIONS_DATA
            REVIEWS_OBJECT_DETECTION_STATS
            REVIEWS_CLASSIFICATION_STATS
            REVIEWS_LABEL_DISTRIBUTION_STATS

            AGGREGATED_LABEL_DISTRIBUTION
            AGGREGATED_OBJECT_DETECTION_STATS
            AGGREGATED_PREDICTIONS_DATA
            AGGREGATED_DRIFTING_PREDICTIONS

        For aggregation, computation may not have been done by the past.
        You will need to force computation of these aggregations and retrieve them again.


        Examples:
            ```python
            my_deployment.get_stats(ServiceMetrics.PREDICTIONS_DATA)
            my_deployment.get_stats(ServiceMetrics.AGGREGATED_DRIFTING_PREDICTIONS, since=3600)
            my_deployment.get_stats(ServiceMetrics.AGGREGATED_LABEL_DISTRIBUTION, model_version=my_model)
            ```
        Arguments:
            service (str): service queried
            model_version (ModelVersion, optional): Model that shall be used when retrieving data.
                Defaults to None.
            from_timestamp (float, optional): System will only retrieve prediction data after this timestamp.
                Defaults to None.
            to_timestamp (float, optional): System will only retrieve prediction data before this timestamp.
                Defaults to None.
            since (int, optional): System will only retrieve prediction data that are in the last seconds.
                Defaults to None.
            includes (List[str], optional): Research will include these ids and excludes others.
                Defaults to None.
            excludes (List[str], optional): Research will exclude these ids.
                Defaults to None.
            tags (List[str], optional): Research will be done filtering by tags.
                Defaults to None.

        Returns:
            A dict with queried statistics about the service you asked
        """
        query_filter = self._build_filter(
            service=service.service,
            model_version=model_version,
            from_timestamp=from_timestamp,
            to_timestamp=to_timestamp,
            since=since,
            includes=includes,
            excludes=excludes,
            tags=tags,
        )

        if service.is_aggregation:
            resp = self.oracle_connexion.get(
                path=f"/api/deployment/{self.id}/stats", params=query_filter
            ).json()
            if "infos" in resp and "info" in resp["infos"]:
                logger.info("This computation is outdated or has never been done.")
                logger.info(
                    "You can compute it again by calling launch_computation with exactly the same params."
                )
            return resp
        else:
            return self.oracle_connexion.get(
                path=f"/api/deployment/{self.id}/predictions/stats",
                params=query_filter,
            ).json()

    @staticmethod
    def _build_filter(
        service: str,
        model_version: Optional[ModelVersion] = None,
        from_timestamp: Optional[float] = None,
        to_timestamp: Optional[float] = None,
        since: Optional[int] = None,
        includes: Optional[List[str]] = None,
        excludes: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
    ) -> dict:
        query_filter = {"service": service}

        if model_version is not None:
            query_filter["model_id"] = model_version.id

        if from_timestamp is not None:
            query_filter["from_timestamp"] = from_timestamp

        if to_timestamp is not None:
            query_filter["to_timestamp"] = to_timestamp

        if since is not None:
            query_filter["since"] = since

        if includes is not None:
            query_filter["includes"] = includes

        if excludes is not None:
            query_filter["excludes"] = excludes

        if tags is not None:
            query_filter["tags"] = tags

        return query_filter

    @exception_handler
    @beartype
    def monitor(
        self,
        image_path: Union[str, Path],
        latency: float,
        height: int,
        width: int,
        prediction: PredictionFormat,
        source: Optional[str] = None,
        tags: Optional[List[str]] = None,
        timestamp: Optional[float] = None,
        model_version: Optional[ModelVersion] = None,
        shadow_model_version: Optional[ModelVersion] = None,
        shadow_latency: Optional[float] = None,
        shadow_raw_predictions: Optional[PredictionFormat] = None,
        shadow_prediction: Optional[PredictionFormat] = None,
        content_type: Optional[Union[SupportedContentType, str]] = None,
    ) -> dict:
        """Send a prediction for this deployment on our monitoring service.

        :warning: Signature of this method has been recently changed and can break some methods :
        - model_version and shadow_model_version are not used anymore : system will use what's currently being monitored in this deployment
        - shadow_raw_predictions has been renamed to shadow_prediction

        Args:
            image_path (str or Path): image path
            latency (float): latency used by model to compute your prediction
            height (int): height of image
            width (int): width of image
            prediction (PredictionFormat): data of your prediction, can be a Classification, a Segmentation or an ObjectDetection Format.
                DetectionPredictionFormat, ClassificationPredictionFormat and SegmentationPredictionFormat:
                    detection_classes (List[int]): list of classes
                    detection_scores (List[float]): list of scores of predictions
                DetectionPredictionFormat and SegmentationPredictionFormat:
                    detection_boxes (List[List[int]]): list of bboxes representing rectangles of your shapes. bboxes are formatted as
                                                            [top, left, bottom, right]
                SegmentationPredictionFormat:
                    detection_masks (List[List[int]]): list of polygons of your shapes. each polygon is a list of points with coordinates flattened
                                                            [x1, y1, x2, y2, x3, y3, x4, y4, ..]

            source (str, optional): source that can give some metadata to your prediction. Defaults to None.
            tags (list of str, optional): tags that can give some metadata to your prediction. Defaults to None.
            timestamp (float, optional): timestamp of your prediction. Defaults to timestamp of monitoring service on reception.
            shadow_latency (float, optional): latency used by shadow model to compute prediction
            shadow_prediction (PredictionFormat, optional): data of your prediction made by shadow model.
            content_type (str, optional): if given, we won't try to infer content type with mimetype library

        Returns:
            a dict of data returned by our monitoring service
        """
        if model_version:  # pragma: no cover
            logger.warning(
                "'model_version' will soon be removed. It is not used anymore"
            )

        if shadow_model_version:  # pragma: no cover
            logger.warning(
                "'shadow_model_version' will soon be removed. It is not used anymore"
            )

        if shadow_raw_predictions and not shadow_prediction:  # pragma: no cover
            logger.warning(
                "'shadow_raw_predictions' parameter will soon be removed. Please use 'shadow_prediction'"
            )
            shadow_prediction = shadow_raw_predictions

        if not content_type:
            content_type = mimetypes.guess_type(image_path, strict=False)[0]
            if content_type is None:  # pragma: no cover
                raise ContentTypeUnknown(
                    f"Content type of {image_path} could not be inferred"
                )

        filename = image_path.split("/")[-1]

        # Open image, encode it into base64, read filename and content type.
        with open(image_path, "rb") as img_file:
            raw_image = img_file.read()

        return self.monitor_bytes(
            raw_image,
            content_type,
            filename,
            latency,
            height,
            width,
            prediction,
            source,
            tags,
            timestamp,
            shadow_latency,
            shadow_prediction,
        )

    @exception_handler
    @beartype
    def monitor_bytes(
        self,
        raw_image: bytes,
        content_type: Union[SupportedContentType, str],
        filename: str,
        latency: float,
        height: int,
        width: int,
        prediction: PredictionFormat,
        source: Optional[str] = None,
        tags: Optional[List[str]] = None,
        timestamp: Optional[float] = None,
        shadow_latency: Optional[float] = None,
        shadow_prediction: Optional[PredictionFormat] = None,
    ) -> dict:
        """Send a prediction for this deployment on our monitoring service.
        You can use this method instead of monitor() if you have a bytes image and not an image file.
        We will convert it into base 64 as utf8 string and send it to the monitoring service.

        Args:
            raw_image (bytes): raw image in bytes
            content_type (Union[SupportedContentType, str]): content type of image, only 'image/jpeg' or 'image/png' currently supported
            filename (str): filename of image
            latency (float): latency used by model to compute your prediction
            height (int): height of image
            width (int): width of image
            prediction (PredictionFormat): data of your prediction, can be a Classification, a Segmentation or an ObjectDetection Format.
                DetectionPredictionFormat, ClassificationPredictionFormat and SegmentationPredictionFormat:
                    detection_classes (List[int]): list of classes
                    detection_scores (List[float]): list of scores of predictions
                DetectionPredictionFormat and SegmentationPredictionFormat:
                    detection_boxes (List[List[int]]): list of bboxes representing rectangles of your shapes. bboxes are formatted as
                                                            [top, left, bottom, right]
                SegmentationPredictionFormat:
                    detection_masks (List[List[int]]): list of polygons of your shapes. each polygon is a list of points with coordinates flattened
                                                            [x1, y1, x2, y2, x3, y3, x4, y4, ..]

            source (str, optional): source that can give some metadata to your prediction. Defaults to None.
            tags (list of str, optional): tags that can give some metadata to your prediction. Defaults to None.
            timestamp (float, optional): timestamp of your prediction. Defaults to timestamp of monitoring service on reception.
            shadow_latency (float, optional): latency used by shadow model to compute prediction
            shadow_prediction (PredictionFormat, optional): data of your prediction made by shadow model.

        Returns:
            a dict of data returned by our monitoring service
        """
        if prediction.model_type != self.type:
            raise BadRequestError(
                f"Prediction shape of this type {prediction.model_type} cannot be used with this model {self.type}"
            )

        try:
            content_type = SupportedContentType.validate(content_type)
        except TypeError:
            raise ContentTypeUnknown(
                f"Content type {content_type} is not supported : {SupportedContentType.values()}"
            )

        # Convert bytes into a base 64 string
        encoded_image = base64.b64encode(raw_image).decode("utf-8")

        payload = {
            "filename": filename,
            "content_type": content_type.value,
            "height": height,
            "width": width,
            "image": encoded_image,
            "raw_predictions": prediction.dict(),
            "latency": latency,
        }

        if source is not None:
            payload["source"] = source

        if tags is not None:
            payload["tags"] = tags

        if timestamp is not None:
            payload["timestamp"] = timestamp

        if shadow_prediction is not None:
            if shadow_latency is None:
                raise BadRequestError(
                    "Shadow latency and shadow raw predictions shall be defined if you want to push a shadow result"
                )
            if shadow_prediction.model_type != self.type:
                raise BadRequestError(
                    f"Prediction shape of this type {prediction.model_type} cannot be used with this model {self.type}"
                )

            payload["shadow_latency"] = shadow_latency
            payload["shadow_raw_predictions"] = shadow_prediction.dict()

        resp = self.oracle_connexion.post(
            path=f"/api/deployment/{self.id}/predictions",
            data=orjson.dumps(payload),
        )

        if resp.status_code != 201:  # pragma: no cover
            raise MonitorError(
                f"Our monitoring service could not handle your prediction: {resp.status_code}. Check {resp.text}"
            )

        return resp.json()



import logging
import os
import warnings
from functools import partial
from typing import List, Optional, Union
from uuid import UUID

import orjson
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

import picsellia
from picsellia import utils
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import PicselliaError
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.datalake import Datalake
from picsellia.sdk.dataset import Dataset
from picsellia.sdk.dataset_version import DatasetVersion
from picsellia.sdk.datasource import DataSource
from picsellia.sdk.deployment import Deployment
from picsellia.sdk.experiment import Experiment
from picsellia.sdk.job import Job
from picsellia.sdk.model import Model
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.project import Project
from picsellia.sdk.tag import Tag
from picsellia.services.datasource import DataSourceService
from picsellia.types.enums import Framework, InferenceType, TagTarget
from picsellia.types.schemas import OrganizationSchema
from picsellia.utils import filter_payload

warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)

logger = logging.getLogger("picsellia")


class Client:
    """Picsellia SDK Client can be used to communicate with Picsellia services.
    You need an API Token, available on web platform in Profile Settings > Token.
    From this client object, you will be able to retrieve python objects representing your data in the platform.

    Examples:
        ```python
        client = Client(api_token="a0c9f3bbf7e8bc175494fc44bfc6f89aae3eb9d0")
        ```

    Arguments:
        api_token (str, optional): your api token, that can be found on the web platform.
            Defaults to None, client will try to find it into your environment variable as 'PICSELLIA_TOKEN'
        organization_id: (str or UUID, optional): Specify an organization to connect to by giving its id.
            Defaults to None, you will be connected to your main Organization.
        organization_name: (str, optional): Specify an organization to connect to by giving its name.
            If id is also given, client will use organization_id.
            Defaults to None, you will be connected to your main Organization.
        host (str, optional): Define a custom host used for platform.
            Defaults to our Picsellia environment "https://app.picsellia.com".
    """

    def __init__(
        self,
        api_token: Optional[str] = None,
        organization_id: Union[str, UUID, None] = None,
        organization_name: Optional[str] = None,
        host: str = "https://app.picsellia.com",
    ):
        if api_token is None:
            if "PICSELLIA_TOKEN" in os.environ:
                token = os.environ["PICSELLIA_TOKEN"]
            else:
                raise Exception(
                    "Please set up the PICSELLIA_TOKEN environment variable or specify your token"
                )
        else:
            token = api_token

        self.connexion = Connexion(host, token)

        # Ping platform to get username and version matching api_token
        try:
            ping_response = self.connexion.get("/sdk/home/ping").json()
        except Exception as e:  # pragma: no cover
            raise PicselliaError(
                "Cannot connect to the platform. Please check api_token, organization and host given.\n"
                f"Error is : {e}"
            )

        sdk_version = ping_response["sdk_version"]

        if sdk_version != picsellia.__version__:  # pragma: no cover
            logger.warning(
                f"\033[93mYou are using an outdated version of the picsellia package ({picsellia.__version__})\033[0m"
            )
            logger.warning(
                f"\033[93mPlease consider upgrading to {sdk_version} with pip install picsellia --upgrade\033[0m"
            )

        if organization_id is not None:
            if isinstance(organization_id, str):
                organization_id = UUID(organization_id)
            self._id = organization_id
            self.sync()
        elif organization_name is not None:
            r = self.connexion.get(
                path="/sdk/organizations/find", params={"name": organization_name}
            ).json()
            self._id = r["id"]
            self.sync()
        else:
            self._id = ping_response["default_organization"]
            self.sync()

        if self._id == ping_response["default_organization"]:
            message = "your"
        else:  # pragma: no cover
            message = self.name + "'s"

        username = ping_response["username"]
        logger.info(f"Hi {Colors.BLUE}{username}{Colors.ENDC}, welcome back. 🥑")
        logger.info(f"Workspace: {Colors.YELLOW}{message}{Colors.ENDC} organization.")

    def __str__(self) -> str:
        return f"Client initialized for organization `{self.name}`"

    def __enter__(self):
        self.connexion = self.connexion.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connexion.__exit__(exc_type, exc_val, exc_tb)

    @property
    def id(self) -> UUID:
        """Organization UUID connected with this (Client)"""
        return self._id

    @property
    def name(self) -> str:
        """Organization name connected with this (Client)"""
        return self._name

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/organization/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> OrganizationSchema:
        schema = OrganizationSchema(**data)
        self._name = schema.name
        if schema.default_connector_id is not None:
            self.connexion.connector_id = schema.default_connector_id
            self.connexion.organization_id = schema.id
        return schema

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(foo_dataset.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/organization/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/organization/{self.id}"

    @exception_handler
    @beartype
    def get_datalake(self) -> Datalake:
        """Retrieve default datalake of this organization

        Examples:
            ```python
            datalake = client.get_datalake()
            ```

        Returns:
            The (Datalake) of the client that you are using
        """
        # Retrieve default datalake
        r = self.connexion.get(f"/sdk/organization/{self.id}/datalake").json()

        return Datalake(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def list_datalakes(self) -> List[Datalake]:
        """Retrieve all datalakes linked to this organization

        Examples:
            ```python
            datalakes = client.list_datalakes()
            ```

        Returns:
            List of (Datalake) of the client that you are using
        """
        # Retrieve default datalake
        r = self.connexion.get(f"/sdk/organization/{self.id}/datalakes").json()
        return list(map(partial(Datalake, self.connexion, self.id), r["items"]))

    @exception_handler
    @beartype
    def create_dataset(
        self,
        name: str,
        description: str = "",
        private: bool = True,
    ) -> Dataset:
        """Create a (Dataset) in this organization.

        This method allows user to create a dataset into the organization currently connected.
        A dataset can then be versioned into (DatasetVersion).
        User can specify name of the dataset, a description and if the dataset is private or not.

        Examples:
            Create a dataset named datatest with data from datalake and version it
            ```python
            foo_dataset = client.create_dataset('foo_dataset')
            foo_dataset_version_1 = foo_dataset.create_version('first')
            some_data = client.get_datalake().list_data(limit=10)
            foo_dataset_version_1.add_data(some_data)
            ```

        Arguments:
            name (str): Name of the dataset. It must be unique in the organization.
            description (str, optional): A description of the dataset. Defaults to ''.
            private (bool, optional): Specify if the dataset is private. Defaults to True.

        Returns:
            A (Dataset) that you can manipulate, connected to Picsellia
        """
        assert name != "", "Dataset name can't be empty"

        payload = {"name": name, "description": description, "private": private}
        r = self.connexion.post(
            f"/sdk/organization/{self.id}/datasets",
            data=orjson.dumps(payload),
        ).json()
        return Dataset(self.connexion, r)

    @exception_handler
    @beartype
    def get_dataset(self, name: str) -> Dataset:
        """Retrieve a dataset by its name

        Examples:
            ```python
            foo_dataset = client.get_dataset('datatest')
            foo_dataset_version = foo_dataset.get_version('first')
            ```

        Arguments:
            name (str): Name of the dataset

        Returns:
            A (Dataset) that you can use and manipulate
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/datasets/find",
            params=params,
        ).json()
        return Dataset(self.connexion, r)

    @exception_handler
    @beartype
    def get_dataset_by_id(self, id: Union[UUID, str]) -> Dataset:
        """Get a dataset by its id

        Examples:
            ```python
            dataset = client.get_dataset('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str): id of the dataset to retrieve

        Returns:
            A (Dataset) that you can use and manipulate
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/datasets/find", params=params
        ).json()
        return Dataset(self.connexion, r)

    @exception_handler
    @beartype
    def list_datasets(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[Dataset]:
        """Retrieve all dataset of current organization

        Examples:
            ```python
            datasets = client.list_datasets()
            ```

        Returns:
            A list of (Dataset) object that belongs to your organization
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/datasets", params=params
        ).json()
        return list(map(partial(Dataset, self.connexion), r["items"]))

    @exception_handler
    @beartype
    def get_dataset_version_by_id(self, id: Union[UUID, str]) -> DatasetVersion:
        """Get a dataset version by its id

        Examples:
            ```python
            dataset_version = client.get_dataset_version_by_id('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str or UUID): id of the dataset version to retrieve

        Returns:
            A (DatasetVersion)
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/datasetversions/find", params=params
        ).json()
        return DatasetVersion(self.connexion, r)

    @exception_handler
    @beartype
    def create_model(
        self,
        name: str,
        type: Union[str, InferenceType] = None,
        framework: Union[str, Framework] = None,
        private: bool = True,
        description: str = "A brand new model!",
    ) -> Model:
        """Create a new model.

        Arguments:
            name (str): Model name to create.
            type (InferenceType): Model type can only be one of (classification, detection, segmentation).
            framework (Framework): Framework used by this model.
            private (bool): A public model can be seen by everyone. Defaults to true.
            description (str): Description of this model


        Returns:
            A (Model) object that you can manipulate
        """
        if type:
            logging.warning(
                "'type' parameter is deprecated and will be removed in future versions. "
                "If you want to give a type to your model version, call update() on ModelVersion."
            )

        if framework:
            logging.warning(
                "'framework' parameter is deprecated and will be removed in future versions. "
                "If you want to give a framework to your model version, call update() on your ModelVersion."
            )

        payload = {
            "name": name,
            "private": private,
            "description": description,
        }
        r = self.connexion.post(
            f"/sdk/organization/{self.id}/models",
            data=orjson.dumps(payload),
        ).json()
        created_model = Model(self.connexion, r)
        logger.info(
            f"📚 Model {created_model.name} created\n📊 🌐 Platform url: {created_model.get_resource_url_on_platform()}"
        )
        return created_model

    @exception_handler
    @beartype
    def get_model_by_id(self, id: Union[UUID, str]) -> Model:
        """Retrieve a model by its id

        Examples:
            ```python
            model = client.get_model_by_id(UUID("d8fae655-5c34-4a0a-a59a-e49c89f20998"))
            ```
        Arguments:
            id (str): id of the model that you are looking for

        Returns:
            A (Model) object that you can manipulate
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/models/find", params=params
        ).json()
        return Model(self.connexion, r)

    def _do_get_model_by_name(self, url: str, name: str):
        params = {"name": name}
        r = self.connexion.get(url, params=params).json()
        return Model(self.connexion, r)

    @exception_handler
    @beartype
    def get_model(self, name: str) -> Model:
        """Retrieve a model by its name.

        Examples:
            ```python
            model = client.get_model("foo_model")
            ```
        Arguments:
            name (str): name of the model you are looking for

        Returns:
            A (Model) object that you can manipulate
        """
        return self._do_get_model_by_name(
            f"/sdk/organization/{self.id}/models/find", name
        )

    @exception_handler
    @beartype
    def get_public_model(self, name: str) -> Model:
        """Retrieve a public model by its name.
           It can only retrieve *public* model.

        Examples:
            ```python
            model = client.get_public_model("foo_public_model")
            ```
        Arguments:
            name (str): name of the public model you are looking for

        Returns:
            A (Model) object that you can manipulate
        """
        return self._do_get_model_by_name("/sdk/public/models/find", name)

    def _do_list_models(self, url: str, params: dict):
        params = filter_payload(params)
        r = self.connexion.get(url, params=params).json()
        return list(map(partial(Model, self.connexion), r["items"]))

    @exception_handler
    @beartype
    def list_models(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[Model]:
        """List all models stored in this organization

        This will return all the models stored
        If no project is found, will throw a ResourceNotFoundError

        Examples:
            ```python
            models = client.list_models()
            ```

        Returns:
            A list of all (Model) that belong to this organization
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        return self._do_list_models(
            f"/sdk/organization/{self.id}/models", params=params
        )

    @exception_handler
    @beartype
    def list_public_models(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
        name: Optional[str] = None,
        type: Union[str, InferenceType, None] = None,
    ) -> List[Model]:
        """List all public models of Picsellia Hub
        This will return all public models of the hub

        Arguments:
            limit (int, optional): Limit number of public models to retrieve.
                Defaults to None, all public models will be retrieved.
            offset (int, optional): Offset to begin with when listing public models.
                Defaults to None, starting at 0.
            order_by (List[str], optional): Some fields to order models against.
                Defaults to None, models will not be sorted
            name (str, optional): A name to filter public models. It will return models with name *containing*
                this parameter. Defaults to None, models will not be filtered

        Examples:
            ```python
            public_detection_models = client.list_public_models(name="yolo")
            ```

        Returns:
            A list of all public (Model) objects
        """
        if type:
            logging.warning(
                "'type' parameter is deprecated and will be removed in future versions."
            )

        params = {
            "limit": limit,
            "offset": offset,
            "order_by": order_by,
            "name": name,
        }
        return self._do_list_models("/sdk/public/models", params=params)

    @exception_handler
    @beartype
    def get_model_version_by_id(self, id: Union[UUID, str]) -> ModelVersion:
        """Get a model version by its id

        Examples:
            ```python
            model_version = client.get_model_version_by_id('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str or UUID): id of the model version to retrieve

        Returns:
            A (ModelVersion)
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/modelversions/find", params=params
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def create_project(
        self, name: str, description: Optional[str] = None, private: bool = True
    ) -> Project:
        """Create a project with given name and parameters

        This project will be registered into used organization.

        Examples:
            ```python
            my_project = client.create_project("my_project", description="My first project!")
            ```
        Arguments:
            name (str): name of the project
            description (str): description of the project
            private (bool): shall be true if its only visible by the contributor of the project

        Returns:
            A (Project) that you can manipulate to run experiments, or attach dataset
        """
        payload = {"name": name, "private": private}
        if description is not None:
            payload["description"] = description

        r = self.connexion.post(
            f"/sdk/organization/{self.id}/projects",
            data=orjson.dumps(payload),
        ).json()
        created_project = Project(self.connexion, r)
        logger.info(
            f"📚 Project {name} created\n📊 Description: {description if description is not None else ''}\n🌐 Platform url: {created_project.get_resource_url_on_platform()}"
        )
        return created_project

    @exception_handler
    @beartype
    def get_project(self, project_name: str) -> Project:
        """Get a project from its name

        Retrieve a project from its name.
        Project must belong to used organization.

        Examples:
            ```python
            my_project = client.get_project("my_project")
            ```
        Arguments:
            project_name (str): name of the project to retrieve

        Returns:
            A (Project) of your organization, you can manipulate to run experiments, or attach dataset
        """
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/projects/find",
            params={"name": project_name},
        ).json()
        return Project(self.connexion, r)

    @exception_handler
    @beartype
    def get_project_by_id(self, id: Union[UUID, str]) -> Project:
        """Get a project from its id

        Retrieve a project from its id.
        Project must belong to used organization.
        If no project is found, will throw a ResourceNotFoundError

        Examples:
            ```python
            my_project = client.get_project("2214aacc-b884-41e1-b70f-420c0cd7eefb")
            ```
        Arguments:
            id (str): id of the project to retrieve

        Returns:
            A (Project) of your organization, you can manipulate to run experiments, or attach dataset
        """
        if isinstance(id, str):
            id = UUID(id)
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/projects/find?id={id}"
        ).json()
        return Project(self.connexion, r)

    @exception_handler
    @beartype
    def list_projects(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[Project]:
        """List all projects of your organization.

        Retrieve all projects of your organization

        Examples:
            ```python
            projects = client.list_projects()
            ```

        Returns:
            A list of Project of your organization
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/projects", params=params
        ).json()
        return list(map(partial(Project, self.connexion), r["items"]))

    @exception_handler
    @beartype
    def get_experiment_by_id(self, id: Union[UUID, str]) -> Experiment:
        """Get an experiment by its id

        Examples:
            ```python
            experiment = client.get_experiment_by_id('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str or UUID): id of the experiment to retrieve

        Returns:
            A (Experiment)
        """
        if isinstance(id, str):
            id = UUID(id)

        params = {"id": id}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/experiments/find", params=params
        ).json()
        return Experiment(self.connexion, r)

    @exception_handler
    @beartype
    def get_deployment(self, name: str) -> Deployment:
        """Get a (Deployment) from its name.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            ```
        Arguments:
            name (str): auto-generated name of your deployment.

        Returns:
            A (Deployment) object connected and authenticated to all the services.
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/deployments/find", params=params
        ).json()
        return Deployment(self.connexion, r)

    @exception_handler
    @beartype
    def get_deployment_by_id(self, id: Union[UUID, str]) -> Deployment:
        """Get a (Deployment) from its name.

        Examples:
            ```python
            deployment = client.get_deployment_id(
                id="YOUR DEPLOYMENT ID"
            )
            ```
        Arguments:
            id (str): deployment id displayed in your deployment settings.

        Returns:
            A (Deployment) object connected and authenticated to all the services.
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/deployments/find", params=params
        ).json()
        return Deployment(self.connexion, r)

    @exception_handler
    @beartype
    def list_deployments(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[Deployment]:
        """List all (Deployment) of your organization

        Examples:
            ```python
            our_deployments = client.list_deployments()
            ```
        Arguments:
            limit (int, optional): number max of results to return
            offset (int, optional): offset of page for pagination
            order_by (list[str], optional): keys on which deployments shall be sorted

        Returns:
            List of (Deployment): all deployments object connected and authenticated to all the services.
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/deployments", params=params
        ).json()
        return list(map(partial(Deployment, self.connexion), r["items"]))

    @exception_handler
    @beartype
    def create_deployment(
        self,
        model_version: ModelVersion,
        shadow_model_version: Optional[ModelVersion] = None,
        name: Optional[str] = None,
        min_threshold: Optional[float] = None,
        target_datalake: Optional[Datalake] = None,
        disable_serving: Optional[bool] = False,
    ):
        """Create a (Deployment) from a model.

        This method allows you to create a (Deployment) on Picsellia. You will then have
        access to the monitoring dashboard and eventually a hosted endpoint.

        Examples:
            Create a serverless deployment
            ```python
            model = client.get_model(name="my-awesome-model")
            deployment = model.create_deployment()
            ```

        Arguments:
            model_version (ModelVersion): ModelVersion to deploy.
            shadow_model_version (ModelVersion, optional): ModelVersion to perform shadow predictions.
            name (str, optional): Name of your deployment. Defaults to a random name.
            min_threshold (float, optional): Threshold of detection scores used by models when predicting. Defaults to 0.
            target_datalake (Datalake, optional): Datalake to use when data are pushed into Picsellia.
                Defaults to organization default Datalake.
            disable_serving (bool, optional): Whether to not use Picsellia Serving. Defaults to False.


        Returns:
            A (Deployment)
        """
        from picsellia.sdk.deployment import Deployment

        payload = {
            "model_version_id": model_version.id,
            "name": name,
            "min_threshold": min_threshold,
        }
        if target_datalake is not None:
            payload["target_datalake_id"] = target_datalake.id
        if shadow_model_version:
            payload["shadow_model_version_id"] = shadow_model_version.id
        if disable_serving:
            payload["disable_serving"] = disable_serving

        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.post(
            f"/sdk/organization/{self.id}/deployments",
            data=orjson.dumps(filtered_payload),
        ).json()
        deployment = Deployment(self.connexion, r)
        logger.info(f"{model_version} is deployed on {deployment}")
        return deployment

    @exception_handler
    @beartype
    def create_datasource(self, name: str) -> DataSource:
        """Create a data source into this organization

        Examples:
            ```python
            data_source = client.create_datasource()
            ```

        Returns:
            A  (DataSource) object that belongs to your organization
        """
        return DataSourceService.create_datasource(self.connexion, self._id, name)

    @exception_handler
    @beartype
    def get_datasource(self, name: str) -> DataSource:
        """Retrieve all data source of current organization

        Examples:
            ```python
            data_sources = client.list_datasources()
            ```

        Returns:
            A list of (DataSource) object that belongs to your organization
        """
        return DataSourceService.get_datasource(self.connexion, self._id, name)

    @exception_handler
    @beartype
    def get_or_create_datasource(self, name: str) -> DataSource:
        """Retrieve a datasource by its name.
        If tag datasource not exist, create it and return it.

        Examples:
            ```python
            tag = self.get_or_create_datasource("new_source")
            ```
        Arguments:
            name (str): Datasource name to retrieve or create

        Returns:
            A (Tag) object
        """
        return DataSourceService.get_or_create_datasource(
            self.connexion, self._id, name
        )

    @exception_handler
    @beartype
    def list_datasources(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[DataSource]:
        """Retrieve all data source of current organization

        Examples:
            ```python
            data_sources = client.list_datasources()
            ```

        Returns:
            A list of (DataSource) object that belongs to your organization
        """
        return DataSourceService.list_datasources(
            self.connexion, self._id, limit, offset, order_by
        )

    @exception_handler
    @beartype
    def _create_tag_organization_scoped(self, name: str, target: TagTarget) -> Tag:
        payload = {"name": name, "target_type": target}
        r = self.connexion.post(
            f"/sdk/organization/{self.id}/tags",
            data=orjson.dumps(payload),
        ).json()
        return Tag(self.connexion, r)

    @exception_handler
    @beartype
    def create_dataset_tag(self, name: str) -> Tag:
        """Create a Dataset Tag, usable only on Dataset objects

        Examples:
            ```python
            tag = organization.create_dataset_tag("global")
            ```
        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.DATASET)

    @exception_handler
    @beartype
    def create_dataset_version_tag(self, name: str) -> Tag:
        """Create a Dataset Version Tag, usable only on Dataset Version objects

        Examples:
            ```python
            tag = organization.create_dataset_version_tag("train")
            ```
        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.DATASET_VERSION)

    @exception_handler
    @beartype
    def create_model_tag(self, name: str) -> Tag:
        """Create a Model Tag, usable only on Model objects

        Examples:
            ```python
            tag = organization.create_model_tag("model")
            ```
        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.MODEL)

    @exception_handler
    @beartype
    def create_model_version_tag(self, name: str) -> Tag:
        """Create a Model Version Tag, usable only on Model Version objects

        Examples:
            ```python
            tag = organization.create_model_version_tag("initial")
            ```
        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.MODEL_VERSION)

    @exception_handler
    @beartype
    def create_deployment_tag(self, name: str) -> Tag:
        """Create a Deployment Tag, usable only on Deployment objects

        Examples:
            ```python
            tag = organization.create_deployment_tag("operation")
            ```
        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.DEPLOYMENT)

    @exception_handler
    @beartype
    def _list_tags_organization_scoped(self, target: TagTarget) -> List[Tag]:
        params = {"target_type": target.value}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/tags", params=params
        ).json()
        return list(map(partial(Tag, self.connexion), r["items"]))

    @exception_handler
    @beartype
    def list_dataset_tags(self) -> List[Tag]:
        """List all Dataset tags, usable only on Dataset objects

        Examples:
            ```python
            tags = organization.list_dataset_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.DATASET)

    @exception_handler
    @beartype
    def list_dataset_version_tags(self) -> List[Tag]:
        """List all Dataset Version tags, usable only on Dataset Version objects

        Examples:
            ```python
            tags = organization.list_dataset_version_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.DATASET_VERSION)

    @exception_handler
    @beartype
    def list_model_tags(self) -> List[Tag]:
        """List all Model tags, usable only on Model objects

        Examples:
            ```python
            tags = organization.list_model_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.MODEL)

    @exception_handler
    @beartype
    def list_model_version_tags(self) -> List[Tag]:
        """List all Model Version tags, usable only on Model Version objects

        Examples:
            ```python
            tags = organization.list_model_version_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.MODEL_VERSION)

    @exception_handler
    @beartype
    def list_deployment_tags(self) -> List[Tag]:
        """List all Deployment tags, usable only on Deployment Version objects

        Examples:
            ```python
            tags = organization.list_deployment_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.DEPLOYMENT)

    @exception_handler
    @beartype
    def _find_tag_organization_scoped(self, target: TagTarget, name: str) -> Tag:
        params = {"name": name, "target_type": target.value}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/tags/find",
            params=params,
        ).json()
        return Tag(self.connexion, r)

    @exception_handler
    @beartype
    def find_dataset_tag(self, name: str) -> Tag:
        """Find a Dataset tag, usable only on Dataset objects, from its name

        Examples:
            ```python
            tag = organization.find_dataset_tag("global")
            ```

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.DATASET, name)

    @exception_handler
    @beartype
    def find_dataset_version_tag(self, name: str) -> Tag:
        """Find a Dataset Version tag, usable only on Dataset Version objects, from its name

        Examples:
            ```python
            tag = organization.find_dataset_version_tag("train")
            ```

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.DATASET_VERSION, name)

    @exception_handler
    @beartype
    def find_model_tag(self, name: str) -> Tag:
        """Find a Model tag, usable only on Model objects, from its name

        Examples:
            ```python
            tag = organization.find_dataset_version_tag("model")
            ```

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.MODEL, name)

    @exception_handler
    @beartype
    def find_model_version_tag(self, name: str) -> Tag:
        """Find a Model Version tag, usable only on Model version objects, from its name

        Examples:
            ```python
            tag = organization.find_model_version_tag("initial")
            ```

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.MODEL_VERSION, name)

    @exception_handler
    @beartype
    def find_deployment_tag(self, name: str) -> Tag:
        """Find a Deployment tag, usable only on Deployment objects, from its name

        Examples:
            ```python
            tag = organization.find_deployment_tag("operation")
            ```

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.DEPLOYMENT, name)

    @exception_handler
    @beartype
    def get_job_by_id(self, id: Union[UUID, str]) -> Job:
        """Get a (Job) from its id.

        Examples:
            ```python
            deployment = client.get_job_by_id(
                id="YOUR JOB ID"
            )
            ```
        Arguments:
            id (str): deployment id displayed in your deployment settings.

        Returns:
            A (Job) object .
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/sdk/organization/{self.id}/jobv2s/find", params=params
        ).json()
        return Job(self.connexion, r, version=2)



import logging
import warnings
from functools import partial
from typing import Dict, List, Optional, Tuple, Union
from uuid import UUID

import orjson
from beartype import beartype
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

import picsellia.pxl_multithreading as mlt
from picsellia import exceptions
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.classification import Classification
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.label import Label
from picsellia.sdk.line import Line
from picsellia.sdk.multi_object import MultiObject
from picsellia.sdk.point import Point
from picsellia.sdk.polygon import Polygon
from picsellia.sdk.rectangle import Rectangle
from picsellia.sdk.worker import Worker
from picsellia.types.enums import AnnotationStatus
from picsellia.types.schemas import AnnotationSchema
from picsellia.utils import chunk_list

logger = logging.getLogger("picsellia")
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)


class Annotation(Dao):
    def __init__(
        self, connexion: Connexion, dataset_version_id: UUID, asset_id: UUID, data: dict
    ) -> None:
        Dao.__init__(self, connexion, data)
        self._dataset_version_id = dataset_version_id
        self._asset_id = asset_id

    @property
    def asset_id(self) -> UUID:
        """UUID of the (Asset) holding this (Annotation)"""
        return self._asset_id

    @property
    def dataset_version_id(self) -> UUID:
        """UUID of (DatasetVersion) holding this (Annotation)"""
        return self._dataset_version_id

    @property
    def duration(self) -> float:
        """Duration time of this (Annotation)"""
        return self._duration

    @property
    def status(self) -> AnnotationStatus:
        """Status of this (Annotation)"""
        return self._status

    @property
    def worker_id(self) -> UUID:
        """UUID of the (Worker) that created this (Annotation)"""
        return self._worker_id

    def __str__(self):
        return f"{Colors.BLUE}Annotation from worker {self.worker_id} on asset {self.asset_id} {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/annotation/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> AnnotationSchema:
        schema = AnnotationSchema(**data)
        self._worker_id = schema.worker_id
        self._duration = schema.duration
        self._status = schema.status
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        worker: Optional[Worker] = None,
        duration: Union[float, int, None] = None,
        status: Union[AnnotationStatus, str, None] = None,
    ) -> None:
        """Update this annotation with a new worker, a new duration or a new status.

        Examples:
            ```python
            one_annotation.update(status=AnnotationStatus.ACCEPTED)
            ```
        Arguments:
            worker (Worker, optional): Worker of this annotation. Defaults to None.
            duration (float, optional): Duration of this annotation. Defaults to None.
            status (AnnotationStatus, optional): Status of this annotation. Defaults to None.
        """
        payload = {}
        if duration:
            payload["duration"] = float(duration)

        if worker:
            payload["worker_id"] = worker.id

        if status:
            payload["status"] = AnnotationStatus.validate(status)

        r = self.connexion.patch(
            f"/sdk/annotation/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this annotation from the platform.
        All annotations shapes will be deleted!
        This is a very dangerous move.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            one_annotation.delete()
            ```
        """
        self.connexion.delete(f"/sdk/annotation/{self.id}")
        logger.info(f"{self} deleted from platform.")

    @exception_handler
    @beartype
    def overwrite(
        self,
        rectangles: Optional[List[Tuple[int, int, int, int, Label]]] = None,
        polygons: Optional[List[Tuple[List[List[int]], Label]]] = None,
        classifications: Optional[List[Label]] = None,
        lines: Optional[List[Tuple[List[List[int]], Label]]] = None,
        points: Optional[List[Tuple[List[int], int, Label]]] = None,
        duration: Union[float, int] = 0.0,
    ) -> None:
        """Overwrite content of this annotation with a new duration and new shapes

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            one_annotation.overwrite(rectangles=[(10, 20, 30, 40, label_cat), (50, 60, 20, 30, label_dog)])
            ```
        """
        payload = {"duration": duration}
        if rectangles:
            payload["rectangles"] = [
                {
                    "x": rectangle[0],
                    "y": rectangle[1],
                    "w": rectangle[2],
                    "h": rectangle[3],
                    "label_id": rectangle[4].id,
                }
                for rectangle in rectangles
            ]

        if polygons:
            payload["polygons"] = [
                {"polygon": polygon[0], "label_id": polygon[1].id}
                for polygon in polygons
            ]
        if lines:
            payload["lines"] = [
                {"line": line[0], "label_id": line[1].id} for line in lines
            ]
        if points:
            payload["points"] = [
                {"point": point[0], "order": point[1], "label_id": point[2].id}
                for point in points
            ]
        if classifications:
            payload["classifications"] = [
                {"label_id": classification.id} for classification in classifications
            ]

        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/save", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(
            f"{self} is now saved with a new content. Old shapes already retrieved are not consistent anymore."
        )

    @exception_handler
    @beartype
    def list_rectangles(self) -> List[Rectangle]:
        """Retrieve all rectangles of this annotation

        Examples:
            ```python
            rects = one_annotation.list_rectangles()
            ```
        """
        r = self.connexion.get(f"/sdk/annotation/{self.id}/rectangles").json()
        return list(map(partial(Rectangle, self.connexion, self.id), r["items"]))

    @exception_handler
    @beartype
    def create_rectangle(
        self, x: int, y: int, w: int, h: int, label: Label
    ) -> Rectangle:
        """Create a rectangle into this annotation.

        Examples:
            ```python
            rect = one_annotation.create_rectangle(420, 69, 100, 10, label_car)
            ```
        """
        payload = {"x": x, "y": y, "w": w, "h": h, "label_id": label.id}
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/rectangles", data=orjson.dumps(payload)
        ).json()
        return Rectangle(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def create_multiple_rectangles(
        self, rectangles: List[Tuple[int, int, int, int, Label]]
    ) -> List[Rectangle]:
        """Create some rectangles into this annotation.

        Examples:
            ```python
            rects = one_annotation.create_multiple_rectangles([(420, 69, 100, 10, label_car), (123, 456, 20, 20, label_person)])
            ```
        """
        assert len(rectangles) > 0, "Please give at least one rectangle"
        payload = []
        for rectangle in rectangles:
            payload.append(
                {
                    "x": rectangle[0],
                    "y": rectangle[1],
                    "w": rectangle[2],
                    "h": rectangle[3],
                    "label_id": rectangle[4].id,
                }
            )
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/rectangles", data=orjson.dumps(payload)
        ).json()
        return list(map(partial(Rectangle, self.connexion, self.id), r))

    @exception_handler
    @beartype
    def list_polygons(self) -> List[Polygon]:
        """Retrieve all polygons of this annotation

        Examples:
            ```python
            polys = one_annotation.list_polygons()
            ```
        """
        r = self.connexion.get(f"/sdk/annotation/{self.id}/polygons").json()
        return list(map(partial(Polygon, self.connexion, self.id), r["items"]))

    @exception_handler
    @beartype
    def create_polygon(self, coords: List[List[int]], label: Label) -> Polygon:
        """Create a polygon into this annotation.

        Examples:
            ```python
            poly = one_annotation.create_polygon([[0, 0], [0, 1], [1, 1], [0, 0]], label_car)
            ```
        """
        payload = {"polygon": coords, "label_id": label.id}
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/polygons", data=orjson.dumps(payload)
        ).json()
        return Polygon(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def create_multiple_polygons(
        self, polygons: List[Tuple[List[List[int]], Label]]
    ) -> List[Polygon]:
        """Create some polygons into this annotation.

        Examples:
            ```python
            polys = one_annotation.create_multiple_polygons([([[0, 0], [0, 1], [1, 1], [0, 0]] label_car), ([[0, 2], [0, 3], [1, 2], [0, 2]] label_person)])
            ```
        """
        assert len(polygons) > 0, "Please give at least one polygon"
        payload = []
        for polygon in polygons:
            payload.append({"polygon": polygon[0], "label_id": polygon[1].id})
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/polygons", data=orjson.dumps(payload)
        ).json()
        return list(map(partial(Polygon, self.connexion, self.id), r))

    @exception_handler
    @beartype
    def list_lines(self) -> List[Line]:
        """Retrieve all lines of this annotation

        Examples:
            ```python
            lines = one_annotation.list_lines()
            ```
        """
        r = self.connexion.get(f"/sdk/annotation/{self.id}/lines").json()
        return list(map(partial(Line, self.connexion, self.id), r["items"]))

    @exception_handler
    @beartype
    def create_line(self, coords: List[List[int]], label: Label) -> Line:
        """Create a line into this annotation.

        Examples:
            ```python
            line = one_annotation.create_line([[0, 0], [0, 1], [1, 1], [0, 0]], label_car)
            ```
        """
        payload = {"line": coords, "label_id": label.id}
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/lines", data=orjson.dumps(payload)
        ).json()
        return Line(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def create_multiple_lines(
        self, lines: List[Tuple[List[List[int]], Label]]
    ) -> List[Line]:
        """Create some lines into this annotation.

        Examples:
            ```python
            lines = one_annotation.create_multiple_lines([([[0, 0], [0, 1], [1, 1]] label_car), ([[0, 2], [0, 3], [1, 2]] label_person)])
            ```
        """
        assert len(lines) > 0, "Please give at least one line"
        payload = []
        for line in lines:
            payload.append({"line": line[0], "label_id": line[1].id})
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/lines", data=orjson.dumps(payload)
        ).json()
        return list(map(partial(Line, self.connexion, self.id), r))

    @exception_handler
    @beartype
    def list_points(self) -> List[Point]:
        """Retrieve all points of this annotation

        Examples:
            ```python
            points = one_annotation.list_points()
            ```
        """
        r = self.connexion.get(f"/sdk/annotation/{self.id}/points").json()
        return list(map(partial(Point, self.connexion, self.id), r["items"]))

    @exception_handler
    @beartype
    def create_point(self, coords: List[int], order: int, label: Label) -> Point:
        """Create a point into this annotation.

        Examples:
            ```python
            poly = one_annotation.create_point([0, 0], 1, label_car)
            ```
        """
        payload = {"point": coords, "order": order, "label_id": label.id}
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/points", data=orjson.dumps(payload)
        ).json()
        return Point(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def create_multiple_points(
        self, points: List[Tuple[List[int], int, Label]]
    ) -> List[Point]:
        """Create some points into this annotation.

        Examples:
            ```python
            polys = one_annotation.create_multiple_points([([0, 0], label_car), ([0, 2], label_person)])
            ```
        """
        assert len(points) > 0, "Please give at least one point"
        payload = []
        for point in points:
            payload.append(
                {"point": point[0], "order": point[1], "label_id": point[2].id}
            )
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/points", data=orjson.dumps(payload)
        ).json()
        return list(map(partial(Point, self.connexion, self.id), r))

    @exception_handler
    @beartype
    def list_classifications(self) -> List[Classification]:
        """Retrieve all classifications of this annotation

        Examples:
            ```python
            classifications = one_annotation.list_classifications()
            ```
        """
        r = self.connexion.get(f"/sdk/annotation/{self.id}/classifications").json()
        return list(map(partial(Classification, self.connexion, self.id), r["items"]))

    @exception_handler
    @beartype
    def create_classification(self, label: Label) -> Classification:
        """Create a classification into this annotation.

        Examples:
            ```python
            classifications = one_annotation.create_classification(label_car)
            ```
        """
        payload = {"label_id": label.id}
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/classifications",
            data=orjson.dumps(payload),
        ).json()
        return Classification(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def create_multiple_classifications(
        self, classifications: List[Label]
    ) -> List[Classification]:
        """Create some classifications into this annotation.

        Examples:
            ```python
            polys = one_annotation.create_multiple_classifications([label_car, label_person])
            ```
        """
        assert len(classifications) > 0, "Please give at least one polygon"
        payload = []
        for classification in classifications:
            payload.append({"label_id": classification.id})
        r = self.connexion.post(
            f"/sdk/annotation/{self.id}/classifications",
            data=orjson.dumps(payload),
        ).json()
        return list(map(partial(Classification, self.connexion, self.id), r))


class MultiAnnotation(MultiObject[Annotation]):
    @beartype
    def __init__(
        self, connexion: Connexion, dataset_version_id: UUID, items: List[Annotation]
    ):
        MultiObject.__init__(self, connexion, items)
        self._dataset_version_id = dataset_version_id

    @property
    def dataset_version_id(self) -> UUID:
        return self._dataset_version_id

    def __str__(self) -> str:
        return f"{Colors.GREEN}MultiAnnotation for dataset version {self.dataset_version_id}{Colors.ENDC}  size: {len(self)}"

    def __getitem__(self, key) -> Union[Annotation, "MultiAnnotation"]:
        if isinstance(key, slice):
            indices = range(*key.indices(len(self.items)))
            annotations = [self.items[i] for i in indices]
            return MultiAnnotation(self.connexion, self.dataset_version_id, annotations)
        return self.items[key]

    @beartype
    def __add__(self, other) -> "MultiAnnotation":
        self.assert_same_connexion(other)
        items = self.items.copy()
        self._add_other_items_to_items(other, items)
        return MultiAnnotation(self.connexion, self.dataset_version_id, items)

    @beartype
    def __iadd__(self, other) -> "MultiAnnotation":
        self.assert_same_connexion(other)
        self._add_other_items_to_items(other, self.items)
        return self

    def _add_other_items_to_items(self, other, items) -> None:
        if isinstance(other, MultiAnnotation):
            if other.dataset_version_id != self.dataset_version_id:
                raise exceptions.BadRequestError(
                    "These annotations does not come from the same dataset"
                )
            items.extend(other.items.copy())
        elif isinstance(other, Annotation):
            if other.dataset_version_id != self.dataset_version_id:
                raise exceptions.BadRequestError(
                    "This annotation does not come from the same dataset"
                )
            items.append(other)
        else:
            raise exceptions.BadRequestError("You can't add these two objects")

    def copy(self) -> "MultiAnnotation":
        return MultiAnnotation(
            self.connexion, self.dataset_version_id, self.items.copy()
        )

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete these annotations and their shape

        :warning: **DANGER ZONE**: Be very careful here!

        Remove these annotations

        Examples:
            ```python
            some_annotations = dataset.list_annotations()[:10]
                some_annotations.delete()
            ```
        """
        payload = {"asset_ids": ["__all__"], "annotation_ids": self.ids}
        self.connexion.delete(
            f"/sdk/dataset/version/{self.dataset_version_id}/annotations",
            data=orjson.dumps(payload),
        )
        logger.info(f"{len(self.items)} annotations deleted.")

    @exception_handler
    @beartype
    def load(
        self,
        chunk_size: int = 1000,
        max_workers: Optional[int] = None,
        skip_error: bool = False,
    ) -> dict:
        """Load these annotation by retrieving shapes with labels, asset_id and worker_id

        Examples:
            ```python
            annotations = foo_dataset.list_annotations()
                annotations_dict = annotations.load()
            ```
        Arguments:
            chunk_size (int, optional): Size of chunk of annotations to load by request. Defaults to 1000.
            max_workers (int, optional): Number of max workers used to load annotations. Defaults to os.cpu_count() + 4.
            skip_error (bool, optional): skip error of a chunk and return partial annotations. Default to False
        """
        if chunk_size < 1 or chunk_size > 10000:
            raise exceptions.BadRequestError(
                "Impossible to load less than 1 or more than 10000 annotations by chunk. Please give another chunk_size"
            )

        return MultiAnnotation.load_annotations_from_ids(
            self.connexion,
            self.dataset_version_id,
            self.ids,
            chunk_size,
            max_workers,
            skip_error,
        )

    @staticmethod
    def load_annotations_from_ids(
        connexion: Connexion,
        dataset_version_id: UUID,
        ids: List[UUID],
        chunk_size: int,
        max_workers: Optional[int] = None,
        skip_error: bool = False,
    ) -> dict:
        def load_some_annotations(list_index: tuple):
            r = connexion.xget(
                f"/sdk/dataset/version/{dataset_version_id}/annotations",
                data=orjson.dumps(list_index[0]),
            )
            return r.json()

        list_chunked = chunk_list(ids, chunk_size)

        results: Dict[int, List[dict]] = mlt.do_mlt_function(
            list_chunked,
            load_some_annotations,
            lambda item: item[1],
            max_workers=max_workers,
        )

        assets: Dict[str, list] = {}
        for key, chunk_annotation in results.items():
            if chunk_annotation is None:
                if skip_error:
                    logging.error(
                        f"Something wrong happened while loading annotations with chunk {key}."
                    )
                    logging.error(
                        "This dataset may not be ready for loading annotations (check its type). Skipping chunk."
                    )
                    continue
                else:
                    raise exceptions.BadRequestError(
                        f"Something wrong happened while loading annotations with chunk {key}."
                        "This dataset may not be ready for loading annotations (check its type)."
                    )

            for annotation in chunk_annotation:
                asset_id = annotation.pop("asset_id")
                if asset_id not in assets:
                    assets[asset_id] = []
                assets[asset_id].append(annotation)

        return assets



import logging
import os
import warnings
from pathlib import Path
from typing import Dict, List, Union
from uuid import UUID

import yaml
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

from picsellia.exceptions import (
    FileNotFoundException,
    ResourceConflictError,
    ResourceNotFoundError,
    UnparsableAnnotationFileException,
)
from picsellia.sdk.asset import Asset, MultiAsset
from picsellia.sdk.label import Label
from picsellia.types.enums import InferenceType

warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
logger = logging.getLogger("picsellia")


def read_filenames_from_file_paths(
    file_paths: List[Union[str, Path]]
) -> Dict[str, Path]:
    files: Dict[str, Path] = {}
    for file_path in file_paths:
        if isinstance(file_path, str):
            file_path = Path(file_path)

        if not os.path.isfile(file_path) and file_path.suffix != "txt":
            raise UnparsableAnnotationFileException(
                f"{file_path} is not a yolo annotation file."
            )

        filename = Path(file_path).name[:-4]  # remove extension
        files[filename] = file_path

    return files


def parse_configuration_file(configuration_yaml_path: Union[str, Path]) -> List[str]:
    try:
        with open(configuration_yaml_path, "r") as yaml_file:
            configuration_yolo = yaml.safe_load(yaml_file)

        nc = configuration_yolo["nc"]
        names = configuration_yolo["names"]

        if not isinstance(nc, int) or not isinstance(names, list):
            raise TypeError(
                "Parameters 'nc' should be an integer and 'names' a list of string"
            )

        if nc != len(names):
            raise ValueError(
                f"Parameter 'nc' is {nc} but there are {len(names)} classes in 'names'"
            )
    except FileNotFoundError:
        raise FileNotFoundException(f"{configuration_yaml_path} not found")
    except KeyError as e:
        raise UnparsableAnnotationFileException(
            f"Could not find key `{e}` inside yolo configuration file {configuration_yaml_path}"
        )

    return names


def match_assets_with_filenames(files: Dict[str, Path], multi_assets: MultiAsset):
    assets: Dict[str, Asset] = {}
    for asset in multi_assets:
        filename = asset.filename[:-4]
        if filename not in files:
            # Retrieved an asset that we were not looking for
            continue

        if filename in assets:  # pragma: no cover
            raise ResourceConflictError(
                f"A filename was retrieved twice in your dataset. Please remove one of it. Prefix of filename is {filename}"
            )

        assets[filename] = asset

    return assets


def assert_coherence_files_assets(
    files: Dict[str, Path], assets: Dict[str, Asset], fail_on_asset_not_found: bool
):
    if fail_on_asset_not_found and len(assets) != len(files):
        assets_not_found = set(files.keys()).difference(set(assets.keys()))
        raise ResourceNotFoundError(
            f"Filenames {assets_not_found} were not found in your dataset"
        )


def parse_files_to_annotations(
    inference_type: InferenceType,
    files: Dict[str, Path],
    assets: Dict[str, Asset],
    labels: List[Label],
) -> List[Dict]:
    annotations: List[Dict] = []
    for filename, asset in assets.items():
        raw_annotations: List[List[str]] = _open_yolo_file(files[filename])
        if len(raw_annotations) == 0:
            logger.info(f"No annotation found in file {filename}")
            continue

        if len(raw_annotations) > 1 and inference_type == InferenceType.CLASSIFICATION:
            raise UnparsableAnnotationFileException(
                f"Could not parse yolo file {filename} : only one classification can be given by file"
            )

        try:
            annotations.append(
                _parse_file_to_annotation(
                    asset, raw_annotations, inference_type, labels
                )
            )
        except Exception as e:
            raise UnparsableAnnotationFileException(
                f"Could not parse yolo annotation file {filename}"
            ) from e

    return annotations


def _open_yolo_file(file_path: Path) -> List[List[str]]:
    raw_annotations = []
    with open(file_path, "r") as txt_file:
        for content in txt_file.readlines():
            if content:
                raw_annotations.append(content.strip().split(" "))

    return raw_annotations


def _parse_file_to_annotation(
    asset: Asset,
    raw_annotations: List,
    inference_type: InferenceType,
    labels: List[Label],
) -> dict:
    rectangles = []
    polygons = []
    classifications = []
    for raw_shape in raw_annotations:
        if len(raw_shape) < 1:  # pragma: no cover
            continue

        label_id = _parse_raw_yolo_annotation_label(raw_shape, labels)

        if inference_type == InferenceType.OBJECT_DETECTION:
            rectangles.append(
                _convert_raw_yolo_annotation_to_rectangle(
                    raw_shape, label_id, asset.width, asset.height
                )
            )
        elif inference_type == InferenceType.SEGMENTATION:
            polygons.append(
                _convert_raw_yolo_annotation_to_polygon(
                    raw_shape, label_id, asset.width, asset.height
                )
            )
        elif inference_type == InferenceType.CLASSIFICATION:
            classifications.append(
                _convert_raw_yolo_annotation_to_classification(label_id)
            )
        else:  # pragma: no cover
            raise TypeError(f"{inference_type} is not supported yet")

    return {
        "asset_id": asset.id,
        "rectangles": rectangles,
        "polygons": polygons,
        "classifications": classifications,
    }


def _parse_raw_yolo_annotation_label(
    raw_shape: List[str],
    labels: List[Label],
):
    label_offset = int(raw_shape[0])
    if label_offset > len(labels):
        raise UnparsableAnnotationFileException(
            f"This annotation has label {label_offset} but it is not defined. "
            f"Check your yolo configuration file"
        )
    return labels[label_offset].id


def _convert_raw_yolo_annotation_to_polygon(
    raw_shape: List[str],
    label_id: UUID,
    asset_width: int,
    asset_height: int,
) -> dict:
    shape_values = _convert_raw_shape_to_shape_values(raw_shape[1:])
    if len(shape_values) % 2 != 0 or len(shape_values) < 6:
        raise UnparsableAnnotationFileException(
            f"This polygon cannot be parsed: {shape_values}. It needs at least 3 points to create a polygon."
        )
    coords = [
        [
            int(shape_values[k] * asset_width),
            int(shape_values[k + 1] * asset_height),
        ]
        for k in range(0, len(shape_values), 2)
    ]
    return {"polygon": coords, "label_id": label_id}


def _convert_raw_yolo_annotation_to_rectangle(
    raw_shape: List[str],
    label_id: UUID,
    asset_width: int,
    asset_height: int,
) -> dict:
    shape_values = _convert_raw_shape_to_shape_values(raw_shape[1:])
    if len(shape_values) != 4:
        raise UnparsableAnnotationFileException(
            f"This rectangle cannot be parsed: {raw_shape}. It needs to be formatted as label_id, x, y, w, h"
        )
    # Yolo format give the shape at its center
    # Also, precision can lead to a negative number, which backend won't allow
    w = shape_values[2] * asset_width
    h = shape_values[3] * asset_height
    x = max(0.0, shape_values[0] * asset_width - w / 2)
    y = max(0.0, shape_values[1] * asset_height - h / 2)

    return {
        "x": int(x),
        "y": int(y),
        "w": int(w),
        "h": int(h),
        "label_id": label_id,
    }


def _convert_raw_yolo_annotation_to_classification(label_id: UUID) -> dict:
    return {"label_id": label_id}


def _convert_raw_shape_to_shape_values(raw_shape: List[str]):
    shape_values = []
    for value in raw_shape:
        if value == "":
            continue

        value = float(value)

        if 0 > value or value > 1:
            raise ValueError(f"A value of {raw_shape} is not in [0, 1]")

        shape_values.append(value)

    return shape_values



import logging

from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.downloadable import Downloadable
from picsellia.types.schemas import ModelFileSchema

logger = logging.getLogger("picsellia")


class ModelFile(Dao, Downloadable):
    def __init__(self, connexion: Connexion, data: dict):
        Dao.__init__(self, connexion, data)
        Downloadable.__init__(self)

    @property
    def name(self) -> str:
        """Name of this (ModelFile)"""
        return self._name

    @property
    def object_name(self) -> str:
        """Object name of this (ModelFile)"""
        return self._object_name

    @property
    def filename(self) -> str:
        """Filename of this (ModelFile)"""
        return self._filename

    @property
    def large(self) -> bool:
        """If True, this (ModelFile) is considered having a large size"""
        return self._large

    def __str__(self):
        return (
            f"{Colors.BLUE}Model file named '{self.name}'{Colors.ENDC} (id: {self.id})"
        )

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/sdk/model/file/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> ModelFileSchema:
        schema = ModelFileSchema(**data)
        self._name = schema.name
        self._object_name = schema.object_name
        self._filename = schema.filename
        self._large = schema.large
        self._url = schema.url
        return schema

    @exception_handler
    @beartype
    def reset_url(self) -> str:
        """Reset url property of this ModelFile by calling platform.

        Returns:
            A url as a string of this ModelFile.
        """
        self._url = self.connexion.init_download(self.object_name)
        return self._url

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this file

        Examples:
            ```python
            model_file.delete()
            ```
        """
        self.connexion.delete(f"/sdk/model/file/{self.id}")
        logger.info(f"{self} deleted from platform.")
